import os

import pytest
import torch
from torch import Tensor
import torch.nn.functional as F

from deepali.core import functional as U
from deepali.core import Grid


def test_fill_border():
    image = torch.zeros((2, 1, 7, 5), dtype=torch.float)

    result = U.fill_border(image, margin=1, value=1)
    assert isinstance(result, Tensor)
    assert result.shape == image.shape
    assert result[:, :, :1, :].eq(1).all()
    assert result[:, :, :, :1].eq(1).all()
    assert result[:, :, 6:, :].eq(1).all()
    assert result[:, :, :, 4:].eq(1).all()
    assert result[:, :, 1:6, 1:4].eq(0).all()
    assert result.sum() == 40

    result = U.fill_border(image, margin=(1, 2), value=1)
    assert isinstance(result, Tensor)
    assert result.shape == image.shape
    assert result[:, :, :2, :].eq(1).all()
    assert result[:, :, :, :1].eq(1).all()
    assert result[:, :, 5:, :].eq(1).all()
    assert result[:, :, :, 4:].eq(1).all()
    assert result[:, :, 2:5, 1:4].eq(0).all()
    assert result.sum() == 52

    image = torch.zeros((2, 1, 7, 5, 11), dtype=torch.float)
    result = U.fill_border(image, margin=(3, 1, 2), value=1)
    assert isinstance(result, Tensor)
    assert result.shape == image.shape
    assert result[:, :, :2, :, :].eq(1).all()
    assert result[:, :, :, :1, :].eq(1).all()
    assert result[:, :, :, :, :3].eq(1).all()
    assert result[:, :, 5:, :, :].eq(1).all()
    assert result[:, :, :, 4:, :].eq(1).all()
    assert result[:, :, :, :, 8:].eq(1).all()
    assert result[:, :, 2:5, 1:4, 3:8].eq(0).all()
    assert result.sum() == 680


def test_finite_differences() -> None:
    image = torch.randn((3, 2, 8, 16, 24))

    # Forward difference scheme
    image.requires_grad = True
    deriv = U.finite_differences(image, "x", mode="forward")
    assert isinstance(deriv, Tensor)
    assert deriv.requires_grad is True
    deriv = deriv.detach()
    assert deriv.requires_grad is False
    assert deriv.dtype.is_floating_point
    assert deriv.shape == image.shape

    expected = F.pad(image, (0, 1, 0, 0, 0, 0), mode="replicate")
    expected = expected[..., 1:].sub(expected[..., :-1])
    assert torch.allclose(deriv, expected)

    # Backward difference scheme
    image.requires_grad = False
    deriv = U.finite_differences(image, "y", mode="backward")
    assert isinstance(deriv, Tensor)
    assert deriv.requires_grad is False
    assert deriv.dtype.is_floating_point
    assert deriv.shape == image.shape

    expected = F.pad(image, (0, 0, 1, 0, 0, 0), mode="replicate")
    expected = expected[..., 1:, :].sub(expected[..., :-1, :])
    assert torch.allclose(deriv, expected)

    # Central difference scheme
    image.requires_grad = False
    deriv = U.finite_differences(image, "z", mode="central")
    assert isinstance(deriv, Tensor)
    assert deriv.requires_grad is False
    assert deriv.dtype.is_floating_point
    assert deriv.shape == image.shape

    expected = F.pad(image, (0, 0, 0, 0, 1, 1), mode="replicate")
    expected = expected[:, :, 2:, :, :].sub(expected[:, :, :-2, :, :]).div(2)
    assert torch.allclose(deriv, expected)

    # Forward-central-backward difference scheme
    deriv = U.finite_differences(image, "z", mode="forward_central_backward")
    assert isinstance(deriv, Tensor)
    assert deriv.requires_grad is False
    assert deriv.dtype.is_floating_point
    assert deriv.shape == image.shape

    expected = torch.cat(
        [
            image[:, :, 1:2, :, :].sub(image[:, :, :1, :, :]),
            image[:, :, 2:, :, :].sub(image[:, :, :-2, :, :]).div(2),
            image[:, :, -1:, :, :].sub(image[:, :, -2:-1, :, :]),
        ],
        dim=2,
    )
    assert torch.allclose(deriv, expected)

    # Zero-th order
    assert U.finite_differences(image, "x", mode="central", order=0) is image


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_rand_sample_cuda() -> None:
    if "CUDA_VISIBLE_DEVICES" not in os.environ:
        os.environ["CUDA_VISIBLE_DEVICES"] = "0"

    device = torch.device("cuda:0")
    generator = torch.Generator(device=device).manual_seed(123456789)

    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)

    shape = torch.Size((5, 1, 32, 64, 128))
    data = torch.arange(shape[2:].numel()).reshape((1, 1) + shape[2:]).expand(shape).to(device)
    num_samples = 10

    # Draw unweighted samples with and without replacement
    t_elapsed_1 = 0
    for i in range(5):
        start.record()
        values = U.rand_sample(data, num_samples, mask=None, replacement=False, generator=generator)
        end.record()
        torch.cuda.synchronize()
        if i > 0:
            t_elapsed_1 += start.elapsed_time(end)
    t_elapsed_1 /= 4

    assert not torch.allclose(values[0], values[1])

    t_elapsed_2 = 0
    for i in range(5):
        start.record()
        values = U.rand_sample(data, num_samples, mask=None, replacement=True, generator=generator)
        end.record()
        torch.cuda.synchronize()
        if i > 0:
            t_elapsed_2 += start.elapsed_time(end)
    t_elapsed_2 /= 4

    assert not torch.allclose(values[0], values[1])

    # Compare with using multinomial with an all-one mask
    t_elapsed_3 = 0
    for i in range(5):
        start.record()
        mask = torch.ones((1, 1) + data.shape[2:], device=device)
        values = U.rand_sample(data, num_samples, mask=mask, replacement=False, generator=generator)
        end.record()
        torch.cuda.synchronize()
        if i > 0:
            t_elapsed_3 += start.elapsed_time(end)
    t_elapsed_3 /= 4

    assert not torch.allclose(values[0], values[1])

    t_elapsed_4 = 0
    for i in range(5):
        start.record()
        mask = torch.ones((1, 1) + data.shape[2:], device=device)
        values = U.rand_sample(data, num_samples, mask=mask, replacement=True, generator=generator)
        end.record()
        torch.cuda.synchronize()
        if i > 0:
            t_elapsed_4 += start.elapsed_time(end)
    t_elapsed_4 /= 4

    assert not torch.allclose(values[0], values[1])


def test_sample_image() -> None:
    shape = torch.Size((5, 2, 32, 64, 63))
    image: Tensor = torch.arange(shape.numel())
    image = image.reshape(shape)
    grid = Grid(shape=shape[2:])

    indices = torch.arange(0, grid.numel(), 10)
    voxels = U.unravel_coords(indices.unsqueeze(0), grid.size())
    coords = grid.index_to_cube(voxels)
    assert coords.dtype == grid.dtype
    assert coords.dtype.is_floating_point
    assert coords.shape == (1, len(indices), 3)
    assert coords.min().ge(-1)
    assert coords.max().le(1)

    result = U.sample_image(image, coords, mode="nearest")
    expected = image.flatten(2).index_select(2, indices)
    assert result.eq(expected).all()



import multiprocessing as mp
from psutil import virtual_memory


def cpu_count() -> int:
    r"""Get number of available CPUs.

    If running inside a Linux Docker container, this returns the limit set when creating the container.

    """
    try:
        with open("/sys/fs/cgroup/cpu/cpu.cfs_quota_us") as fp:
            cfs_quota_us = int(fp.read())
        with open("/sys/fs/cgroup/cpu/cpu.cfs_period_us") as fp:
            cfs_period_us = int(fp.read())
        # For physical machine, the `cfs_quota_us` could be '-1'
        num_cpus = cfs_quota_us // cfs_period_us
    except FileNotFoundError:
        num_cpus = -1
    if num_cpus < 1:
        num_cpus = mp.cpu_count()
    return num_cpus


def memory_limit() -> int:
    r"""Get memory limit in bytes.

    If running inside a Linux Docker container, this returns the limit set when creating the container.

    """
    limit_in_bytes = virtual_memory().total
    try:
        with open("/sys/fs/cgroup/memory/memory.limit_in_bytes") as fp:
            limit_in_bytes = min(int(fp.read()), limit_in_bytes)
    except FileNotFoundError:
        pass
    return limit_in_bytes



r"""Auxiliary functions and classes for dealing with configuration files."""

from __future__ import annotations

from dataclasses import asdict, fields
from io import StringIO
from typing import Any, Dict, Mapping, Optional, Sequence, TypeVar, Type

import dacite
import json

# PyYAML does not support YAML 1.2, which is required by DVC for example
# (cf. https://github.com/iterative/dvc/issues/8466#issuecomment-1290757564)
from ruamel.yaml import YAML

from .pathlib import Path, PathUri, abspath_template, is_uri
from .storage import StorageObject
from .typing import is_path_str_field


T = TypeVar("T", bound="DataclassConfig")


class DataclassConfig(object):
    r"""Base class of configuration data classes."""

    @staticmethod
    def section() -> str:
        r"""Common key prefix of configuration entries in configuration file."""
        return ""

    @classmethod
    def _from_dict(cls: Type[T], arg: Mapping[str, Any], parent: Optional[Path] = None) -> T:
        r"""Create configuration from dictionary."""
        config = dacite.from_dict(cls, arg)
        return config

    @classmethod
    def from_dict(cls: Type[T], arg: Mapping[str, Any], parent: Optional[Path] = None) -> T:
        r"""Create configuration from dictionary."""
        config = cls._from_dict(arg, parent=parent)
        config._finalize(parent)
        return config

    @classmethod
    def from_path(cls: Type[T], path: PathUri, section: Optional[str] = None) -> T:
        r"""Load configuration from file."""
        config = read_config_dict(path)
        if section is None:
            section = cls.section()
        if section:
            for key in section.split("."):
                config = config.get(key, {})
        if is_uri(path):
            if path.startswith("file:///"):
                parent = Path(path[7:]).parent
            else:
                parent = None
        else:
            parent = Path(path).absolute().parent
        return cls.from_dict(config, parent=parent)

    @classmethod
    def read(cls: Type[T], path: PathUri, section: Optional[str] = None) -> T:
        r"""Load configuration from JSON or YAML file."""
        return cls.from_path(path, section=section)

    def write(self, path: PathUri) -> None:
        r"""Save configuration to JSON or YAML file."""
        config = self.asdict()
        write_config_dict(config, path)

    def asdict(self) -> Dict[str, Any]:
        r"""Get configuration dictionary."""
        return asdict(self)

    def _finalize(self: T, parent: Optional[Path] = None) -> None:
        r"""Finalize parameters after loading these from input file."""
        for field in fields(self):
            value = getattr(self, field.name)
            if value is None:
                continue
            if isinstance(value, DataclassConfig):
                value._finalize(parent)
            elif is_path_str_field(field):
                value = abspath_template(value, parent=parent)
                setattr(self, field.name, value)

    def _join_kwargs_in_sequence(self, attr: str):
        r"""Merge kwarg dictionaries in 'norm' or 'acti' sequence.

        In case of a sequence instead of a single string or dictionary, ``ConvLayer`` expects
        ``norm`` and ``acti`` parameters to be a sequence of length 2, where the first entry
        is the normalization or activiation layer name, respectively, and the second entry is
        a dictionary of keyword arguments for the respective layer. In a YAML file, it may
        be convenient, however, to specify these arguments on a single line as follows:

        .. code-block:: yaml

            norm: [batch, momentum: 0.1, eps: 0.001]

        This represents a sequence, where the first item is a string and the following items
        are dictionaries with a single key. This ``__post_init__`` function merges these separate
        dictionaries into a single dictionary in order to support above YAML syntax.

        The same functionality may be useful for other configuration entries, not only
        ``normalization()``, ``activiation()``, or ``pooling()`` related parameters.

        Alternatively, one could use a single dictionary in the YAML configuration:

        .. code-block:: yaml

            norm: {name: batch, momentum: 0.1, eps: 0.001}

        Args:
            attr: Name of dataclass attribute to modify in place.

        """
        arg = getattr(self, attr)
        arg = join_kwargs_in_sequence(arg)
        setattr(self, attr, arg)


def join_kwargs_in_sequence(arg):
    r"""Merge kwarg dictionaries in 'norm' or 'acti' sequence.

    In case of a sequence instead of a single string or dictionary, ``ConvLayer`` expects
    ``norm`` and ``acti`` parameters to be a sequence of length 2, where the first entry
    is the normalization or activiation layer name, respectively, and the second entry is
    a dictionary of keyword arguments for the respective layer. In a YAML file, it may
    be convenient, however, to specify these arguments on a single line as follows:

    .. code-block:: yaml

        norm: [batch, momentum: 0.1, eps: 0.001]

    This represents a sequence, where the first item is a string and the following items
    are dictionaries with a single key. This ``__post_init__`` function merges these separate
    dictionaries into a single dictionary in order to support above YAML syntax.

    The same functionality may be useful for other configuration entries, not only
    ``normalization()``, ``activiation()``, or ``pooling()`` related parameters.

    Alternatively, one could use a single dictionary in the YAML configuration:

    .. code-block:: yaml

        norm: {name: batch, momentum: 0.1, eps: 0.001}

    Args:
        attr: Name of dataclass attribute to modify in place.

    """
    if not isinstance(arg, str) and isinstance(arg, Sequence):
        if not all(isinstance(item, (str, dict) if i == 0 else dict) for i, item in enumerate(arg)):
            raise TypeError(
                "join_kwargs_in_sequence() 'arg' must be str, dict,"
                " or sequence of dicts with the first item being either a str or dict"
            )
        if len(arg) == 1 and isinstance(arg[0], dict):
            arg = arg[0]
        elif len(arg) > 1:
            start = 0 if isinstance(arg[0], dict) else 1
            kwargs = dict(arg[start])
            for i in range(start + 1, len(arg)):
                cur = arg[i]
                assert isinstance(cur, dict)
                for name, value in cur.items():
                    if name in kwargs:
                        raise ValueError(
                            f"join_kwargs_in_sequence() 'arg' has duplicate kwarg {name}"
                        )
                    kwargs[name] = value
            arg = kwargs if start == 0 else (arg[0], kwargs)
    return arg


def read_config_dict(path: PathUri) -> Dict[str, Any]:
    r"""Read configuration from JSON or YAML file."""
    with StorageObject.from_path(path) as config_object:
        config_suffix = config_object.path.suffix
        if config_suffix.lower() not in (".json", ".yaml", ".yml"):
            raise ValueError(f"read_config_dict() 'path' has unsupported suffix {config_suffix}")
        config_text = config_object.read_text()
        if config_suffix == ".json":
            config_dict = json.loads(config_text)
        else:
            config_dict = YAML(typ="safe").load(config_text)
    if config_dict is None:
        config_dict = {}
    return config_dict


def write_config_dict(config: Dict[str, Any], path: PathUri) -> None:
    r"""Write configuration to JSON or YAML file."""
    with StorageObject.from_path(path) as config_object:
        config_suffix = config_object.path.suffix
        if config_suffix.lower() not in (".json", ".yaml", ".yml"):
            raise ValueError(f"write_config_dict() 'path' has unsupported suffix {config_suffix}")
        if config_suffix == ".json":
            config_text = json.dumps(config) + "\n"
        else:
            buffer = StringIO()
            YAML(typ="safe").dump(config, buffer)
            config_text = buffer.getvalue()
        config_object.write_text(config_text)



from typing import Mapping, Union

from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics.metric import Metric, reinit__is_reduced, sync_all_reduce

import torch
from torch import Tensor


class AverageLoss(Metric):
    r"""Calculates the average loss given batched loss values stored in ``engine.state.output``.

    This metric can be used instead of ``ignite.metrics.loss.Loss`` when the average loss values
    for each example in the input batch are stored as 1-dimensional tensor in the ``engine.state.output``.
    The (transformed) return value of the engines ``process_function`` must be either a tensor of the
    computed loss values, or a dictionary containing the key ``"loss"``.

    """

    @reinit__is_reduced
    def reset(self) -> None:
        self.accumulator = 0
        self.num_examples = 0

    @reinit__is_reduced
    def update(self, output: Union[Tensor, Mapping]) -> None:
        loss = output.get("loss") if isinstance(output, Mapping) else output
        if not isinstance(loss, Tensor):
            raise TypeError("AverageLoss.update() 'output' loss value must be torch.Tensor")
        if loss.ndim == 0:
            self.accumulator += loss.item()
            self.num_examples += 1
        elif loss.dim() == 1:
            self.accumulator += loss.sum().item()
            self.num_examples += loss.shape[0]
        else:
            raise ValueError(
                "AverageLoss.update() 'output' loss value must be scalar or 1-dimensional tensor"
            )

    @sync_all_reduce("accumulator", "num_examples")
    def compute(self) -> float:
        if self.num_examples == 0:
            raise NotComputableError(
                "Loss must have at least one example before it can be computed."
            )
        return self.accumulator / self.num_examples

    @torch.no_grad()
    def iteration_completed(self, engine: Engine) -> None:
        output = self._output_transform(engine.state.output)
        self.update(output)



r"""Specialized subtypes of ``torch.Tensor``, datasets thereof, and data transforms."""

from .collate import collate_samples
from .prepare import prepare_batch

from .dataset import Dataset
from .dataset import ImageDataset
from .dataset import ImageDatasetConfig
from .dataset import MetaDataset
from .dataset import GroupDataset
from .dataset import JoinDataset

from .partition import Partition
from .partition import dataset_split_lengths

from .image import Image
from .image import ImageBatch

from .flow import FlowField
from .flow import FlowFields


__all__ = (
    # Tensor specializations
    "FlowField",
    "FlowFields",
    "Image",
    "ImageBatch",
    # Data collections
    "Dataset",
    "GroupDataset",
    "ImageDataset",
    "ImageDatasetConfig",
    "JoinDataset",
    "MetaDataset",
    # Data partitioning
    "dataset_split_lengths",
    "Partition",
    # Data loading
    "collate_samples",
    "prepare_batch",
)



r"""Registration loss for pairwise image registration."""

from collections import defaultdict
import re
from typing import Dict, Generator, List, Mapping, Optional, Sequence, Tuple, Type, TypeVar, Union

import torch
from torch import Tensor
from torch.nn import Module

from deepali.core import Grid, PaddingMode, Sampling, functional as U
from deepali.losses import PairwiseImageLoss, PointSetDistance, LandmarkPointDistance
from deepali.losses import BSplineLoss, DisplacementLoss, ParamsLoss
from deepali.losses import RegistrationLoss, RegistrationLosses, RegistrationResult
from deepali.modules import SampleImage
from deepali.spatial import BSplineTransform, SpatialTransform
from deepali.spatial import CompositeTransform, SequentialTransform


RE_WEIGHT = re.compile(
    r"^((?P<mul>[0-9]+(\.[0-9]+)?)\s*[\* ])?\s*(?P<chn>[a-zA-Z0-9_-]+)\s*(\+\s*(?P<add>[0-9]+(\.[0-9]+)?))?$"
)
RE_TERM_VAR = re.compile(r"^[a-zA-Z0-9_-]+\((?P<var>[a-zA-Z0-9_]+)\)$")


TModule = TypeVar("TModule", bound=Module)
TSpatialTransform = TypeVar("TSpatialTransform", bound=SpatialTransform)


class PairwiseImageRegistrationLoss(RegistrationLoss):
    r"""Loss function for pairwise multi-channel image registration."""

    def __init__(
        self,
        source_data: Tensor,
        target_data: Tensor,
        source_grid: Grid,
        target_grid: Grid,
        source_chns: Mapping[str, Union[int, Tuple[int, int]]],
        target_chns: Mapping[str, Union[int, Tuple[int, int]]],
        source_pset: Optional[Tensor] = None,
        target_pset: Optional[Tensor] = None,
        source_landmarks: Optional[Tensor] = None,
        target_landmarks: Optional[Tensor] = None,
        losses: Optional[RegistrationLosses] = None,
        weights: Mapping[str, Union[float, str]] = None,
        transform: Optional[Union[CompositeTransform, SpatialTransform]] = None,
        sampling: Union[Sampling, str] = Sampling.LINEAR,
    ):
        r"""Initialize multi-channel registration loss function.

        Args:
            source_data: Moving normalized multi-channel source image batch tensor.
            source_data: Fixed normalized multi-channel target image batch tensor.
            source_grid: Sampling grid of source image.
            source_grid: Sampling grid of target image.
            source_chns: Mapping from channel (loss, weight) name to index or range.
            target_chns: Mapping from channel (loss, weight) name to index or range.
            source_pset: Point sets defined with respect to source image grid.
            target_pset: Point sets defined with respect to target image grid.
            source_landmarks: Landmark points defined with respect to source image grid.
            target_landmarks: Landmark points defined with respect to target image grid.
            losses: Dictionary of named loss terms. Loss terms must be either a subclass of
                ``PairwiseImageLoss``, ``DisplacementLoss``, ``PointSetDistance``, ``ParamsLoss``,
                or ``torch.nn.Module``. In case of a ``PairwiseImageLoss``, the key (name) of the
                loss term must be found in ``channels`` which identifies the corresponding ``target``
                and ``source`` data channels that this loss term relates to. If the name is not found
                in the ``channels`` mapping, the loss term is called with all image channels as input.
                If a loss term is not an instance of a known registration loss type, it is assumed to be a
                regularization term without arguments, e.g., a ``torch.nn.Module`` which itself has a reference
                to the parameters of the transformation that it is based on.
            weights: Scalar weights of loss terms or name of channel with locally adaptive weights.
            transform: Spatial transformation to apply to ``source`` image.
            sampling: Image interpolation mode.

        """
        super().__init__()
        self.register_buffer("_source_data", source_data)
        self.register_buffer("_target_data", target_data)
        self.source_grid = source_grid
        self.target_grid = target_grid
        self.source_chns = dict(source_chns or {})
        self.target_chns = dict(target_chns or {})
        self.source_pset = source_pset
        self.target_pset = target_pset
        self.source_landmarks = source_landmarks
        self.target_landmarks = target_landmarks
        if transform is None:
            transform = SequentialTransform(self.target_grid)
        elif isinstance(transform, SpatialTransform):
            transform = SequentialTransform(transform)
        elif not isinstance(transform, CompositeTransform):
            raise TypeError(
                "PairwiseImageRegistrationLoss() 'transform' must be of type CompositeTransform"
            )
        self.transform = transform
        self._sample_image = SampleImage(
            target=self.target_grid,
            source=self.source_grid,
            sampling=sampling,
            padding=PaddingMode.ZEROS,
            align_centers=False,
        )
        points = self.target_grid.coords(device=self._target_data.device)
        self.register_buffer("grid_points", points.unsqueeze(0))
        self.loss_terms = self.as_module_dict(losses)
        self.weights = dict(weights or {})

    @property
    def device(self) -> torch.device:
        r"""Device on which loss is evaluated."""
        device = self._target_data.device
        assert isinstance(device, torch.device)
        return device

    def loss_terms_of_type(self, loss_type: Type[TModule]) -> Dict[str, TModule]:
        r"""Get dictionary of loss terms of a specifictype."""
        return {
            name: module
            for name, module in self.loss_terms.items()
            if isinstance(module, loss_type)
        }

    def transforms_of_type(
        self, transform_type: Type[TSpatialTransform]
    ) -> List[TSpatialTransform]:
        r"""Get list of spatial transformations of a specific type."""

        def _iter_transforms(
            transform,
        ) -> Generator[SpatialTransform, None, None]:
            if isinstance(transform, transform_type):
                yield transform
            elif isinstance(transform, CompositeTransform):
                for t in transform.transforms():
                    yield from _iter_transforms(t)

        transforms = list(_iter_transforms(self.transform))
        return transforms

    @property
    def has_transform(self) -> bool:
        r"""Whether a spatial transformation is set."""
        return len(self.transform) > 0

    def target_data(self) -> Tensor:
        r"""Target image tensor."""
        data = self._target_data
        assert isinstance(data, Tensor)
        return data

    def source_data(self, grid: Optional[Tensor] = None) -> Tensor:
        r"""Sample source image at transformed target grid points."""
        data = self._source_data
        assert isinstance(data, Tensor)
        if grid is None:
            return data
        return self._sample_image(grid, data)

    def data_mask(self, data: Tensor, channels: Dict[str, Union[int, Tuple[int, int]]]) -> Tensor:
        r"""Get boolean mask from data tensor."""
        slice_ = self.as_slice(channels["msk"])
        start, stop = slice_.start, slice_.stop
        mask = data.narrow(1, start, stop - start)
        return mask > 0.9

    def overlap_mask(self, source: Tensor, target: Tensor) -> Optional[Tensor]:
        r"""Overlap mask at which to evaluate pairwise data term."""
        mask = self.data_mask(source, self.source_chns)
        mask &= self.data_mask(target, self.target_chns)
        return mask

    @classmethod
    def as_slice(cls, arg: Union[int, Sequence[int]]) -> slice:
        r"""Slice of image data channels associated with the specified name."""
        if isinstance(arg, int):
            arg = (arg,)
        if len(arg) == 1:
            arg = (arg[0], arg[0] + 1)
        if len(arg) == 2:
            arg = (arg[0], arg[1], 1)
        if len(arg) != 3:
            raise ValueError(
                f"{cls.__name__}.as_slice() 'arg' must be int or sequence of length 1, 2, or 3"
            )
        return slice(*arg)

    @classmethod
    def data_channels(cls, data: Tensor, c: slice) -> Tensor:
        r"""Get subimage data tensor of named channel."""
        i = (slice(0, data.shape[0]), c) + tuple(
            slice(0, data.shape[dim]) for dim in range(2, data.ndim)
        )
        return data[i]

    def loss_input(
        self, name: str, data: Tensor, channels: Dict[str, Union[int, Tuple[int, int]]]
    ) -> Tensor:
        r"""Get input for named loss term."""
        if name in channels:
            c = channels[name]
        elif "img" not in channels:
            raise RuntimeError(f"Channels map contains neither entry for '{name}' nor 'img'")
        else:
            c = channels["img"]
        i: slice = self.as_slice(c)
        return self.data_channels(data, i)

    def loss_mask(
        self,
        name: str,
        data: Tensor,
        channels: Dict[str, Union[int, Tuple[int, int]]],
        mask: Tensor,
    ) -> Tensor:
        r"""Get mask for named loss term."""
        weight = self.weights.get(name, 1.0)
        if not isinstance(weight, str):
            return mask
        match = RE_WEIGHT.match(weight)
        if match is None:
            raise RuntimeError(f"Invalid weight string ('{weight}') for loss term '{name}'")
        chn = match.group("chn")
        mul = match.group("mul")
        add = match.group("add")
        c = channels.get(chn)
        if c is None:
            raise RuntimeError(
                f"Channels map contains no entry for '{name}' weight string '{weight}'"
            )
        i = self.as_slice(c)
        w = self.data_channels(data, i)
        if mul is not None:
            w = w * float(mul)
        if add is not None:
            w = w + float(add)
        return w * mask

    def eval(self) -> RegistrationResult:
        r"""Evaluate pairwise image registration loss."""
        result = {}
        losses = {}
        misc_excl = set()
        # Transform target grid points
        x: Tensor = self.grid_points
        y: Tensor = self.transform(x, grid=True)
        # Buffered vector fields
        variables = defaultdict(list)
        for name, buf in self.transform.named_buffers():
            if buf.requires_grad:
                var = name.rsplit(".", 1)[-1]
                variables[var].append(buf)
        variables["w"] = [U.move_dim(y - x, -1, 1)]
        # Sum of pairwise image dissimilarity terms
        data_terms = self.loss_terms_of_type(PairwiseImageLoss)
        misc_excl |= set(data_terms.keys())
        if data_terms:
            source = self.source_data(y)
            target = self.target_data()
            mask = self.overlap_mask(source, target)
            for name, term in data_terms.items():
                s = self.loss_input(name, source, self.source_chns)
                t = self.loss_input(name, target, self.target_chns)
                m = self.loss_mask(name, target, self.target_chns, mask)
                losses[name] = term(s, t, mask=m)
            result["source"] = source
            result["target"] = target
            result["mask"] = mask
        # Sum of pairwise point set distance terms
        dist_terms = self.loss_terms_of_type(PointSetDistance)
        misc_excl |= set(dist_terms.keys())
        ldist_terms = {k: v for k, v in dist_terms.items() if isinstance(v, LandmarkPointDistance)}
        dist_terms = {k: v for k, v in dist_terms.items() if k not in ldist_terms}
        if dist_terms:
            if self.source_pset is None:
                raise RuntimeError(f"{type(self).__name__}() missing source point set")
            if self.target_pset is None:
                raise RuntimeError(f"{type(self).__name__}() missing target point set")
            s = self.source_pset
            t = self.transform(self.target_pset)
            for name, term in dist_terms.items():
                losses[name] = term(t, s)
        if ldist_terms:
            if self.source_landmarks is None:
                raise RuntimeError(f"{type(self).__name__}() missing source landmarks")
            if self.target_landmarks is None:
                raise RuntimeError(f"{type(self).__name__}() missing target landmarks")
            s = self.source_landmarks
            t = self.transform(self.target_landmarks)
            for name, term in ldist_terms.items():
                losses[name] = term(t, s)
        # Sum of displacement field regularization terms
        disp_terms = self.loss_terms_of_type(DisplacementLoss)
        misc_excl |= set(disp_terms.keys())
        for name, term in disp_terms.items():
            match = RE_TERM_VAR.match(name)
            if match:
                var = match.group("var")
            elif "v" in variables:
                var = "v"
            elif "u" in variables:
                var = "u"
            else:
                var = "w"
            bufs = variables.get(var)
            if not bufs:
                raise RuntimeError(f"Unknown variable in loss term name '{name}'")
            value = torch.tensor(0, dtype=torch.float, device=self.device)
            for buf in bufs:
                value += term(buf)
            losses[name] = value
        # Sum of free-form deformation loss terms
        bspline_transforms = self.transforms_of_type(BSplineTransform)
        bspline_terms = self.loss_terms_of_type(BSplineLoss)
        misc_excl |= set(bspline_terms.keys())
        for name, term in bspline_terms.items():
            value = torch.tensor(0, dtype=torch.float, device=self.device)
            for bspline_transform in bspline_transforms:
                value += term(bspline_transform.data())
            losses[name] = value
        # Sum of parameters loss terms
        params_terms = self.loss_terms_of_type(ParamsLoss)
        misc_excl |= set(params_terms.keys())
        for name, term in params_terms.items():
            value = torch.tensor(0, dtype=torch.float, device=self.device)
            count = 0
            for params in self.transform.parameters():
                value += term(params)
                count += 1
            if count > 1:
                value /= count
            losses[name] = value
        # Sum of other regularization terms
        misc_terms = {k: v for k, v in self.loss_terms.items() if k not in misc_excl}
        for name, term in misc_terms.items():
            losses[name] = term()
        # Calculate total loss
        result["losses"] = losses
        result["weights"] = self.weights
        result["loss"] = self._weighted_sum(losses)
        return result

    def _weighted_sum(self, losses: Mapping[str, Tensor]) -> Tensor:
        r"""Compute weighted sum of loss terms."""
        loss = torch.tensor(0, dtype=torch.float, device=self.device)
        weights = self.weights
        for name, value in losses.items():
            w = weights.get(name, 1.0)
            if not isinstance(w, str):
                value = w * value
            loss += value.sum()
        return loss


def weight_channel_names(weights: Mapping[str, Union[float, str]]) -> Dict[str, str]:
    r"""Get names of channels that are used to weight loss term of another channel."""
    names = {}
    for term, weight in weights.items():
        if not isinstance(weight, str):
            continue
        match = RE_WEIGHT.match(weight)
        if match is None:
            continue
        names[term] = match.group("chn")
    return names



r"""Engine state output transformations for use as checkpoint score function."""

from typing import Mapping

from ignite.engine import Engine

from torch import Tensor

from deepali.core.collections import get_tensor


def negative_loss_score_function(engine: Engine, key: str = "loss") -> Tensor:
    r"""Get negated loss value from ``engine.state.output``."""
    output = engine.state.output
    if isinstance(output, Mapping):
        loss = get_tensor(output, key)
    elif isinstance(output, Tensor):
        loss = output
    else:
        raise ValueError("negative_loss_score_function() engine output loss must be a Tensor")
    loss = loss.detach().sum()
    return -float(loss)



from copy import deepcopy
from typing import Tuple

import numpy as np
import pytest
import torch
from torch import Tensor
from torch.nn import functional as F

from deepali.core import ALIGN_CORNERS, Grid, functional as U
from deepali.data import Image, ImageBatch


def image_size(sdim: int) -> Tuple[int, ...]:
    if sdim == 2:
        return (64, 57)
    if sdim == 3:
        return (64, 57, 31)
    raise ValueError("image_size() 'sdim' must be 2 or 3")


def image_shape(sdim: int) -> Tuple[int, ...]:
    return tuple(reversed(image_size(sdim)))


@pytest.fixture(scope="function")
def grid(request) -> Tensor:
    size = image_size(request.param)
    spacing = (0.25, 0.2, 0.5)[: len(size)]
    return Grid(size=size, spacing=spacing)


@pytest.fixture(scope="function")
def data(request) -> Tensor:
    shape = image_shape(request.param)
    return torch.randn((1,) + shape).mul_(100)


@pytest.fixture(scope="function")
def zeros(request) -> Tensor:
    shape = image_shape(request.param)
    return torch.zeros((1,) + shape)


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_init(zeros: Tensor, grid: Grid) -> None:
    data = zeros

    image = Image(data)
    assert type(image) is Image
    assert hasattr(image, "_grid")
    assert isinstance(image._grid, Grid)
    assert image._grid.size() == grid.size()
    assert image.data_ptr() == data.data_ptr()
    tensor = image.tensor()
    assert type(tensor) is Tensor
    assert image.data_ptr() == tensor.data_ptr()

    image = Image(data, grid, device=data.device)
    assert type(image) is Image
    assert hasattr(image, "_grid")
    assert image._grid is grid
    assert image.data_ptr() == data.data_ptr()

    image = Image(data, grid, requires_grad=True)
    assert type(image) is Image
    assert image._grid is grid
    assert image.requires_grad
    assert not image.is_pinned()
    assert image.device == data.device
    assert image.dtype == data.dtype
    assert image.data_ptr() == data.data_ptr()

    if torch.cuda.is_available():
        device = torch.device("cuda:0")
        image = Image(data, grid, device=device)
        assert type(image) is Image
        assert image._grid is grid
        assert image.data_ptr() != data.data_ptr()
        assert image.is_cuda
        assert image.device == device
        assert image.dtype == data.dtype

        image = Image(data, grid, pin_memory=True)
        assert type(image) is Image
        assert image._grid is grid
        assert not image.requires_grad
        assert image.is_pinned()
        assert image.device == data.device
        assert image.dtype == data.dtype
        assert image.data_ptr() != data.data_ptr()

        image = Image(data, grid, requires_grad=True, pin_memory=True)
        assert type(image) is Image
        assert image._grid is grid
        assert image.requires_grad
        assert image.is_pinned()
        assert image.device == data.device
        assert image.dtype == data.dtype
        assert image.data_ptr() != data.data_ptr()

    params = torch.zeros((1, 32, 64, 64), requires_grad=True)
    assert params.requires_grad

    image = Image(params)
    assert type(image) is Image
    assert image.requires_grad
    assert image.data_ptr() == params.data_ptr()

    image = Image(params, requires_grad=False)
    assert type(image) is Image
    assert not image.requires_grad
    assert image.data_ptr() == params.data_ptr()


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (3,)], indirect=True)
def test_image_deepcopy(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    other = deepcopy(image)
    assert type(other) is Image
    assert hasattr(other, "_grid")
    assert other.grid() is not image.grid()
    assert other.grid() == image.grid()
    assert other.data_ptr() != image.data_ptr()


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (3,)], indirect=True)
def test_image_torch_function(zeros: Tensor, grid: Grid) -> None:
    data = zeros

    image = Image(data, grid)
    assert type(image) is Image
    assert hasattr(image, "_grid")
    assert image.grid() is grid

    result = image.type(image.dtype)
    assert result is image

    result = image.type(torch.int16)
    assert result is not image
    assert type(result) is Image
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.dtype is torch.int16
    assert result.data_ptr() != image.data_ptr()

    result = image.eq(0)
    assert type(result) is Image
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.dtype is torch.bool
    assert result.data_ptr() != image.data_ptr()

    result = result.all()
    assert type(result) is Tensor
    assert not hasattr(result, "_grid")
    assert result.ndim == 0
    assert result.item() is True

    result = torch.add(image, 2)
    assert type(result) is Image
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.eq(2).all()

    result = torch.add(4, image)
    assert type(result) is Image
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.eq(4).all()

    result = image.add(1)
    assert type(result) is Image
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.eq(1).all()

    result = image.clone()
    assert type(result) is Image
    assert hasattr(result, "_grid")
    assert result.grid() is not image.grid()
    assert result.grid() == image.grid()
    assert result.data_ptr() != image.data_ptr()

    result = image.to("cpu", torch.int16)
    assert type(result) is Image
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.device == torch.device("cpu")
    assert result.dtype == torch.int16

    if torch.cuda.is_available():
        result = image.cuda()
        assert type(result) is Image
        assert hasattr(result, "_grid")
        assert result.grid() is image.grid()
        assert result.is_cuda

    image_1 = Image(data, grid)
    image_2 = Image(data.clone(), grid)
    assert image_1.shape[1:] == image_2.shape[1:]
    image_3 = torch.cat([image_1, image_2], dim=0)
    assert isinstance(image_3, Image)
    assert image_3.shape[0] == image_1.shape[0] + image_2.shape[0]
    assert image_3.shape[1:] == image_1.shape[1:]

    # F.grid_sample() should always return a Tensor because the grid of the resampled image
    # may only by chance have the same size as the one on which the data is being resampled,
    # but the spatial positions of the new grid locations differ. See also Image.sample(grid).
    coords = image.grid().coords()
    with pytest.raises(ValueError):
        F.grid_sample(image, coords, align_corners=True)

    result = F.grid_sample(image.batch(), coords.unsqueeze(0), align_corners=True)
    assert type(result) == Tensor
    assert result.shape[1:] == image.shape

    # Split image along channel dimension
    result = image.split(1)
    assert type(result) == tuple
    assert len(result) == 1
    assert type(result[0]) == Image
    assert result[0].grid() is image.grid()

    multi_channel_image = image.repeat((5,) + (1,) * image.grid().ndim)
    assert type(multi_channel_image) == Image
    assert multi_channel_image.shape[0] == 5
    assert multi_channel_image.shape[1:] == image.grid().shape
    assert multi_channel_image.grid() is image.grid()
    result = multi_channel_image.split(2, dim=0)
    assert type(result) == tuple
    assert len(result) == 3
    assert type(result[0]) == Image
    assert type(result[1]) == Image
    assert type(result[2]) == Image
    assert result[0].shape == (2,) + image.shape[1:]
    assert result[1].shape == (2,) + image.shape[1:]
    assert result[2].shape == (1,) + image.shape[1:]

    # Split batch of images along batch dimension
    batch = image.batch()
    assert type(batch) == ImageBatch
    batch.grids() == (image.grid(),)
    batch.grid(0) is image.grid()
    result = batch.tensor_split([0])
    assert type(result) == tuple
    assert len(result) == 2
    assert type(result[0]) == ImageBatch
    assert len(result[0]) == 0
    assert type(result[1]) == ImageBatch
    assert len(result[1]) == 1
    assert result[1].grid(0) is image.grid()

    other: Image = image.clone()
    assert other.grid() is not image.grid()
    batch = torch.cat([image.batch(), other.batch()], dim=0)
    assert type(batch) == ImageBatch
    assert len(batch._grid) == 2
    assert batch.grid(0) is image.grid()
    assert batch.grid(1) is other.grid()

    result = batch.split(1)
    assert type(result) == tuple
    assert len(result) == 2
    assert type(result[0]) == ImageBatch
    assert len(result[0]) == 1
    assert type(result[1]) == ImageBatch
    assert len(result[1]) == 1
    assert result[0].grid(0) is image.grid()
    assert result[1].grid(0) is other.grid()

    result = torch.split(batch, 1)
    assert type(result) == tuple
    assert len(result) == 2
    assert type(result[0]) == ImageBatch
    assert len(result[0]) == 1
    assert type(result[1]) == ImageBatch
    assert len(result[1]) == 1
    assert result[0].grid(0) is image.grid()
    assert result[1].grid(0) is other.grid()


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_batch(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    batch = image.batch()
    assert type(batch) is ImageBatch
    assert hasattr(batch, "_grid")
    assert type(batch._grid) is tuple
    assert len(batch._grid) == 1
    assert batch.data_ptr() == image.data_ptr()


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_batch_getitem(zeros: Tensor, grid: Grid) -> None:
    grids = [deepcopy(grid) for _ in range(5)]  # make grids distinguishable
    batch = ImageBatch(zeros.unsqueeze(0).expand((5,) + zeros.shape), grids, dtype=torch.uint8)
    for i in range(len(batch)):
        batch.tensor()[i] = i
    for i in range(len(batch)):
        image = batch[i]
        assert isinstance(image, Image)
        assert image.grid() is batch.grid(i)
        assert torch.allclose(image, torch.tensor(i, dtype=batch.dtype))

    image = batch[-1]
    assert isinstance(image, Image)
    assert image.grid() is batch.grid(-1)
    assert torch.allclose(image, torch.tensor(len(batch) - 1, dtype=batch.dtype))

    with pytest.raises(IndexError):
        batch[len(batch)]

    other = batch[0:1]
    assert type(other) is ImageBatch
    assert len(other.grids()) == len(other)
    assert len(other) == 1
    assert other.grid(0) is batch.grid(0)
    assert torch.allclose(other, torch.tensor(0, dtype=batch.dtype))

    other = batch[1:-1, 0:1]
    assert type(other) is ImageBatch
    assert len(other.grids()) == len(other)
    assert len(other) == len(batch) - 2
    for i, j in enumerate(range(1, len(batch) - 1)):
        assert other.grid(i) is batch.grid(j)
        assert torch.allclose(other.tensor()[i], torch.tensor(j, dtype=batch.dtype))
    assert torch.allclose(other, batch.tensor()[1:-1, 0:1])

    image = batch[3, :]
    assert isinstance(image, Image)
    assert image.grid() == grid
    assert torch.allclose(image, torch.tensor(3, dtype=image.dtype))

    other = batch[4, 0]
    assert type(other) is Tensor
    assert torch.allclose(other, torch.tensor(4, dtype=image.dtype))
    assert torch.allclose(other, batch.tensor()[4, 0])

    other = batch[(0, 0)]
    assert type(other) is Tensor
    assert torch.allclose(other, batch.tensor()[(0, 0)])

    other = batch[[0, 2]]
    assert type(other) is ImageBatch
    assert other.grids()[0] is batch.grid(0)
    assert other.grids()[1] is batch.grid(2)
    assert torch.allclose(other.tensor(), batch.tensor()[[0, 2]])

    index = np.array([3, 4, 1])
    item = batch[index]
    assert type(item) is ImageBatch
    assert len(item) == 3
    assert item.grid(0) is batch.grid(3)
    assert item.grid(1) is batch.grid(4)
    assert item.grid(2) is batch.grid(1)
    assert torch.allclose(item.tensor(), batch.tensor()[index])

    index = torch.tensor([1, 2])
    item = batch[index]
    assert type(item) is ImageBatch
    assert len(item) == 2
    assert item.grid(0) is batch.grid(1)
    assert item.grid(1) is batch.grid(2)
    assert torch.allclose(item.tensor(), batch.tensor()[index])

    image = batch[4, 0:1]
    assert isinstance(image, Image)
    assert image.nchannels == 1
    assert image.grid() == grid
    assert torch.allclose(image, torch.tensor(4, dtype=image.dtype))

    other = batch[:]
    assert type(other) is type(batch)
    assert other.data_ptr() == batch.data_ptr()

    index = slice(0, 5, 2)
    other = batch[index]
    assert type(other) is type(batch)
    assert other.shape[0] == 3
    assert other.shape[1:] == batch.shape[1:]
    assert other.grid(0) is batch.grid(0)
    assert other.grid(1) is batch.grid(2)
    assert other.grid(2) is batch.grid(4)
    assert torch.allclose(other.tensor(), batch.tensor()[index])

    other = batch[:, :, :, :]
    assert isinstance(other, ImageBatch)
    assert other.grid() == batch.grid()
    assert other.shape == batch.shape
    assert torch.allclose(other, batch)

    other = batch[...]
    assert isinstance(other, ImageBatch)
    assert other.grid() == batch.grid()
    assert other.shape == batch.shape
    assert torch.allclose(other, batch)

    other = batch[:, ...]
    assert isinstance(other, ImageBatch)
    assert other.grid() == batch.grid()
    assert other.shape == batch.shape
    assert torch.allclose(other, batch)

    other = batch[:, :, ...]
    assert isinstance(other, ImageBatch)
    assert other.grid() == batch.grid()
    assert other.shape == batch.shape
    assert torch.allclose(other, batch)

    other = batch[..., :]
    assert isinstance(other, ImageBatch)
    assert other.grid() == batch.grid()
    assert other.shape == batch.shape
    assert torch.allclose(other, batch)

    if batch.ndim == 4:
        other = batch[..., 2, 0:1, :, :]
        assert type(other) is Image
        assert other.grid() is batch.grid(2)
        assert torch.allclose(other, batch.tensor()[..., 2, 0:1, :, :])
        another = batch[2, ..., 0:1, ..., :, :, ...]
        assert type(another) is Image
        assert another.grid() is batch.grid(2)
        assert torch.allclose(another, batch.tensor()[2, ..., 0:1, ..., :, :, ...])
        another = batch[..., 2, ..., 0:1, ..., :, :, ...]
        assert type(another) is Image
        assert torch.allclose(another, batch.tensor()[..., 2, ..., 0, ..., :, :, ...])

    other = batch[:, ..., 0]
    assert type(other) is Tensor

    image = batch[3, ...]
    assert type(image) is Image
    assert image.grid() is batch.grid(3)
    assert torch.allclose(image, batch.tensor()[3, ...])


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (3,)], indirect=True)
def test_image_grid(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    assert image.grid() is grid

    new_grid = Grid(size=grid.size(), spacing=(0.5, 0.4, 0.3))

    new_image = image.grid(new_grid)
    assert new_image is not image
    assert image.grid() is grid
    assert new_image.grid() is new_grid
    assert new_image.data_ptr() == image.data_ptr()

    other_image = image.grid_(new_grid)
    assert other_image is image
    assert image.grid() is new_grid


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (3,)], indirect=True)
def test_image_align_corners(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    assert image.align_corners() == grid.align_corners()
    assert grid.align_corners() == ALIGN_CORNERS
    grid.align_corners_(not ALIGN_CORNERS)
    assert image.align_corners() == grid.align_corners()
    assert grid.align_corners() == (not ALIGN_CORNERS)


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (3,)], indirect=True)
def test_image_center(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    assert isinstance(image.center(), Tensor)
    assert image.center().shape == grid.center().shape
    assert image.center().shape == (grid.ndim,)
    assert torch.allclose(image.center(), grid.center())


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_origin(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    assert isinstance(image.origin(), Tensor)
    assert image.origin().shape == grid.origin().shape
    assert image.origin().shape == (grid.ndim,)
    assert torch.allclose(image.origin(), grid.origin())


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_spacing(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    assert isinstance(image.spacing(), Tensor)
    assert image.spacing().shape == grid.spacing().shape
    assert image.spacing().shape == (grid.ndim,)
    assert torch.allclose(image.spacing(), grid.spacing())


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_direction(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    assert isinstance(image.direction(), Tensor)
    assert image.direction().shape == grid.direction().shape
    assert image.direction().shape == (grid.ndim, grid.ndim)
    assert torch.allclose(image.direction(), grid.direction())


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_sdim(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    assert type(image.sdim) is int
    assert image.sdim == grid.ndim


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_nchannels(zeros: Tensor, grid: Grid) -> None:
    image = Image(zeros, grid)
    assert type(image.nchannels) is int
    assert image.nchannels == 1


@pytest.mark.parametrize("data,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_normalize(data: Tensor, grid: Grid) -> None:
    atol = 1e-7  # assert result.min() == 0 occasionally fails with atol=1e-8

    image = Image(data, grid)
    assert image.min().lt(0)
    assert image.max().gt(1)

    result = image.normalize(mode="unit")
    assert type(result) is Image
    assert result is not image
    assert result.data_ptr() != image.data_ptr()
    assert image.eq(data).all()
    assert torch.allclose(result.min(), torch.tensor(0.0), atol=atol)
    assert torch.allclose(result.max(), torch.tensor(1.0), atol=atol)

    assert image.normalize().eq(result).all()

    result = image.normalize("center")
    assert type(result) is Image
    assert result is not image
    assert result.data_ptr() != image.data_ptr()
    assert image.eq(data).all()
    assert torch.allclose(result.min(), torch.tensor(-0.5), atol=atol)
    assert torch.allclose(result.max(), torch.tensor(0.5), atol=atol)

    result = image.normalize_("center")
    assert type(result) is Image
    assert result is not image
    assert result.data_ptr() == image.data_ptr()
    assert torch.allclose(result.min(), torch.tensor(-0.5), atol=atol)
    assert torch.allclose(result.max(), torch.tensor(0.5), atol=atol)
    assert torch.allclose(image.min(), torch.tensor(-0.5), atol=atol)
    assert torch.allclose(image.max(), torch.tensor(0.5), atol=atol)


@pytest.mark.parametrize("data,grid", [(d, d) for d in (2, 3)], indirect=True)
def test_image_rescale(data: Tensor, grid: Grid) -> None:
    image = Image(data, grid)
    input_min = image.min()
    input_max = image.max()
    assert input_min.lt(0)
    assert input_max.gt(1)

    result = image.rescale(0, 1)
    assert type(result) is Image
    assert result is not image
    assert result.data_ptr() != image.data_ptr()
    assert image.eq(data).all()
    assert torch.allclose(result.min(), torch.tensor(0.0))
    assert torch.allclose(result.max(), torch.tensor(1.0))

    assert torch.allclose(image.rescale(0, 1, input_min, input_max), result)

    result = image.rescale(-0.5, 0.5, data_max=input_max)
    assert type(result) is Image
    assert result is not image
    assert result.data_ptr() != image.data_ptr()
    assert image.eq(data).all()
    assert torch.allclose(result.min(), torch.tensor(-0.5))
    assert torch.allclose(result.max(), torch.tensor(0.5))

    result = image.rescale(0, 255, dtype=torch.uint8)
    assert type(result) is Image
    assert result is not image
    assert result.data_ptr() != image.data_ptr()
    assert result.dtype == torch.uint8
    assert result.min().eq(0)
    assert result.max().eq(255)


def test_image_sample() -> None:
    shape = torch.Size((2, 32, 64, 63))
    data: Tensor = torch.arange(shape.numel())
    data = data.reshape(shape)
    grid = Grid(shape=shape[1:])
    image = Image(data, grid)

    indices = torch.arange(0, grid.numel(), 10)
    voxels = U.unravel_coords(indices, grid.size())
    coords = grid.index_to_cube(voxels)
    assert coords.dtype == grid.dtype
    assert coords.dtype.is_floating_point
    assert coords.shape == (len(indices), grid.ndim)
    assert coords.min().ge(-1)
    assert coords.max().le(1)

    result = image.sample(coords, mode="nearest")
    expected = data.flatten(1).index_select(1, indices)
    assert type(result) is Tensor
    assert result.dtype == image.dtype
    assert result.shape == (image.nchannels, *coords.shape[:-1])
    assert result.shape == expected.shape
    assert result.eq(expected).all()

    result = image.sample(coords, mode="linear")
    assert type(result) is Tensor
    assert result.is_floating_point()
    assert result.dtype != image.dtype

    # Grid points
    coords = grid.coords()
    assert coords.ndim == image.ndim
    result = image.sample(coords, mode="nearest")
    assert type(result) is Tensor
    assert result.dtype == image.dtype
    assert result.device == image.device
    assert result.shape[0] == image.nchannels
    assert result.shape[1:] == grid.shape
    assert torch.allclose(result, image)

    # Batch of grid points
    coords = grid.coords()
    coords = coords.unsqueeze(0).repeat((3,) + (1,) * coords.ndim)
    assert coords.ndim == image.ndim + 1
    result = image.sample(coords, mode="nearest")
    assert type(result) is Tensor
    assert result.dtype == image.dtype
    assert result.device == image.device
    assert result.shape[0] == image.nchannels
    assert result.shape[1] == coords.shape[0]
    assert result.shape[2:] == grid.shape

    batch_result = image.batch().sample(coords, mode="nearest")
    assert type(result) is Tensor
    assert batch_result.dtype == image.dtype
    assert batch_result.device == image.device
    assert batch_result.shape[0] == coords.shape[0]
    assert batch_result.shape[1] == image.nchannels
    assert batch_result.shape[2:] == grid.shape
    assert torch.allclose(result, batch_result.transpose(0, 1))

    # Sampling grid
    result = image.sample(grid)
    assert type(result) is Image
    assert result.grid() is grid
    assert result.data_ptr() == image.data_ptr()
    assert result is image

    shape = torch.Size((3, 31, 63, 63))
    data: Tensor = torch.arange(shape.numel())
    data = data.reshape(shape)
    grid = Grid(shape=shape[1:])
    image = Image(data, grid)

    ds_grid = grid.downsample()
    ds_image = image.sample(ds_grid)
    assert type(ds_image) is Image
    assert ds_image.grid() is ds_grid
    assert ds_image.nchannels == image.nchannels
    assert ds_image.device == image.device
    assert ds_image.is_floating_point()

    indices = torch.arange(0, grid.numel())
    voxels = U.unravel_coords(indices, grid.size())
    coords = grid.index_to_cube(voxels)
    coords = coords.reshape(grid.shape + (grid.ndim,))
    coords = coords[::2, ::2, ::2, :]
    assert torch.allclose(coords, ds_grid.coords())

    result = image.sample(coords)
    assert result.is_floating_point()
    assert torch.allclose(ds_image, result)



r"""Type annotations for torch functions."""

from dataclasses import Field
from pathlib import Path
from typing import Any, Dict, List, Mapping, NamedTuple
from typing import Optional, Protocol, Sequence, Tuple, TypeVar, Union

import torch

from torch import Tensor


EllipsisType = type(...)

T = TypeVar("T")

BoolStr = Union[bool, str]

ScalarOrTuple = Union[T, Tuple[T, ...]]
ScalarOrTuple1d = Union[T, Tuple[T]]
ScalarOrTuple2d = Union[T, Tuple[T, T]]
ScalarOrTuple3d = Union[T, Tuple[T, T, T]]
ScalarOrTuple4d = Union[T, Tuple[T, T, T, T]]
ScalarOrTuple5d = Union[T, Tuple[T, T, T, T, T]]
ScalarOrTuple6d = Union[T, Tuple[T, T, T, T, T, T]]

# Cannot use Sequence type annotation when using torch.jit.ScriptModule
ListOrTuple = Union[List[T], Tuple[T, ...]]

Device = torch.device
DeviceStr = Union[torch.device, str]
DType = torch.dtype
DTypeStr = Union[torch.dtype, str]
Name = Optional[str]  # Optional tensor dimension name
Size = ScalarOrTuple[int]  # Order of spatial dimensions: (X, ...)
Shape = ScalarOrTuple[int]  # Order of spatial dimensions: (..., X)
Scalar = Union[int, float, Tensor]
Array = Union[Sequence[Scalar], Tensor]

PathStr = Union[Path, str]
PathUri = Union[Path, str]


class Dataclass(Protocol):
    r"""Type annotation for any dataclass."""

    __dataclass_fields__: Dict[str, Any]


# While a Batch and Sample have the same type, the difference is in their values
# - Sample: Values must be without batch dimension if tensors/arrays, but can also be of other
#           types such as a file Path or SimpleITK.Image, or None if optional.
# - Batch: Values must be tensors with batch dimension (first dimension) or None if optional.
Batch = Union[Dataclass, Dict[str, Any], NamedTuple]
Sample = Union[Dataclass, Dict[str, Any], NamedTuple]

TensorMapOrSequence = Union[Mapping[str, Tensor], Sequence[Tensor]]

TensorCollection = Union[
    TensorMapOrSequence,
    Mapping[str, TensorMapOrSequence],
    Sequence[TensorMapOrSequence],
]


def is_bool_dtype(dtype: DType) -> bool:
    r"""Checks if ``dtype`` of given NumPy array or PyTorch tensor is boolean type."""
    return dtype in (torch.bool,)


def is_float_dtype(dtype: DType) -> bool:
    r"""Checks if ``dtype`` of given tensor is a floating point type."""
    return dtype in (torch.float16, torch.float32, torch.float64)


def is_int_dtype(dtype: DType) -> bool:
    r"""Checks if ``dtype`` of given tensor is a signed integer type."""
    return dtype in (torch.int8, torch.int16, torch.int32, torch.int64)


def is_uint_dtype(dtype: DType) -> bool:
    r"""Checks if ``dtype`` of given tensor is an unsigned integer type."""
    return dtype in (torch.uint8,)


def is_namedtuple(arg: Any) -> bool:
    r"""Check if given object is a named tuple."""
    return isinstance(arg, tuple) and hasattr(arg, "_fields")


def is_optional_field(field: Field) -> bool:
    r"""Whether given dataclass field type is ``Optional[T] = Union[T, NoneType]``."""
    return is_optional_type_hint(field.type)


def is_optional_type_hint(type_hint: Any) -> bool:
    r"""Whether given type hint is ``Optional[T] = Union[T, NoneType]``."""
    type_origin = getattr(type_hint, "__origin__", None)
    if type_origin is Union:
        return type(None) in type_hint.__args__


def is_path_str(arg: Any) -> bool:
    r"""Whether given object is of type ``pathlib.Path`` or ``str``."""
    return isinstance(arg, (Path, str))


def is_path_str_type_hint(type_hint: Any, required: bool = False) -> bool:
    r"""Check if given type annotation is ``pathlib.Path``, ``PathStr = Union[pathlib.Path, str]``.

    Args:
        type_hint: Type annotation, e.g., ``dataclasses.Field.type``.
        required: Whether path argument is required. If ``False``, ``type(None)`` in the
            type hint is ignore, i.e., also ``Optional[T]`` is considered valid.

    Returns:
        Whether type hint is ``pathlib.Path``, ``Union[pathlib.Path, str]``, or
        ``Union[str, pathlib.Path]``. When ``required=False``, type annotations
        ``Optional[T] = Union[T, None]`` where ``T`` is one of the aforementioned
        path string types also results in a return value of ``True``.

    """
    if type_hint in (Path, "Path", "Optional[Path]", "PathStr", "Optional[PathStr]"):
        return True
    type_origin = getattr(type_hint, "__origin__", None)
    if type_origin is Union:
        type_args = set(type_hint.__args__)
        if not required:
            type_args.discard(type(None))
            type_args.discard("None")
        type_args.discard(str)
        type_args.discard("str")
        if not type_args:
            return False
        for type_arg in type_args:
            if type_arg not in (Path, "Path", "PathStr"):
                return False
        return True
    return False


def is_path_str_field(field: Field, required: bool = False) -> bool:
    r"""Check if given dataclass field type is ``pathlib.Path``, ``PathStr = Union[pathlib.Path, str]``."""
    return is_path_str_type_hint(field.type, required=required)



r"""Utility functions for point sets."""

from typing import Optional, Tuple, Union

import torch
import torch.nn.functional as F
from torch import Tensor

from . import affine as A
from .flow import warp_grid, warp_points
from .grid import ALIGN_CORNERS
from .tensor import move_dim


def bounding_box(points: Tensor) -> Tuple[Tensor, Tensor]:
    r"""Compute corners of minimum axes-aligned bounding box of given points."""
    return points.amin(0), points.amax(0)


def distance_matrix(x: Tensor, y: Tensor) -> Tensor:
    r"""Compute squared Euclidean distances between all pairs of points.

    Args:
        x: Point set of shape ``(N, X, D)``.
        y: Point set of shape ``(N, Y, D)``.

    Returns:
        Tensor of distance matrices of shape ``(N, X, Y)``.

    See also:
        https://discuss.pytorch.org/t/efficient-distance-matrix-computation/9065

    """
    if not isinstance(x, Tensor) or not isinstance(y, Tensor):
        raise TypeError("distance_matrix() 'x' and 'y' must be torch.Tensor")
    if x.ndim != 3 or y.ndim != 3:
        raise ValueError("distance_matrix() 'x' and 'y' must have shape (N, X, D)")
    N, _, D = x.shape
    if y.shape[0] != N:
        raise ValueError("distance_matrix() 'x' and 'y' must have same batch size N")
    if y.shape[2] != D:
        raise ValueError("distance_matrix() 'x' and 'y' must have same point dimension D")
    out_dtype = x.dtype
    if not out_dtype.is_floating_point:
        out_dtype = torch.float32
    x = x.type(torch.float64)
    y = y.type(torch.float64)
    x_norm = x.pow(2).sum(2).view(N, -1, 1)
    y_norm = y.pow(2).sum(2).view(N, 1, -1)
    dist = x_norm + y_norm - 2.0 * torch.bmm(x, torch.transpose(y, 1, 2))
    return dist.type(out_dtype)


def closest_point_distances(x: Tensor, y: Tensor, split_size: int = 10000) -> Tensor:
    r"""Compute minimum Euclidean distance from each point in ``x`` to point set ``y``.

    Args:
        x: Point set of shape ``(N, X, D)``.
        y: Point set of shape ``(N, Y, D)``.
        split_size: Maximum number of points in ``x`` to consider each time when computing
            the full distance matrix between these points in ``x`` and every point in ``y``.
            This is required to limit the size of the distance matrix.

    Returns:
        Tensor of shape ``(N, X)`` with minimum distances from points in ``x`` to points in ``y``.

    """
    if not isinstance(x, Tensor) or not isinstance(y, Tensor):
        raise TypeError("closest_point_distances() 'x' and 'y' must be torch.Tensor")
    if x.ndim != 3 or y.ndim != 3:
        raise ValueError("closest_point_distances() 'x' and 'y' must have shape (N, X, D)")
    N, _, D = x.shape
    if y.shape[0] != N:
        raise ValueError("closest_point_distances() 'x' and 'y' must have same batch size N")
    if y.shape[2] != D:
        raise ValueError("closest_point_distances() 'x' and 'y' must have same point dimension D")
    x = x.float()
    y = y.type(x.dtype)
    min_dists = torch.empty(x.shape[0:2], dtype=x.dtype, device=x.device)
    for i, points in enumerate(x.split(split_size, dim=1)):
        dists = distance_matrix(points, y)
        j = slice(i * split_size, i * split_size + points.shape[1])
        min_dists[:, j] = torch.min(dists, dim=2).values
    return min_dists


def closest_point_indices(x: Tensor, y: Tensor, split_size: int = 10000) -> Tensor:
    r"""Determine indices of points in ``y`` with minimum Euclidean distance from each point in ``x``.

    Args:
        x: Point set of shape ``(N, X, D)``.
        y: Point set of shape ``(N, Y, D)``.
        split_size: Maximum number of points in ``x`` to consider each time when computing
            the full distance matrix between these points in ``x`` and every point in ``y``.
            This is required to limit the size of the distance matrix.

    Returns:
        Tensor of shape ``(N, X)`` with indices of closest points in ``y``.

    """
    if not isinstance(x, Tensor) or not isinstance(y, Tensor):
        raise TypeError("closest_point_indices() 'x' and 'y' must be torch.Tensor")
    if x.ndim != 3 or y.ndim != 3:
        raise ValueError("closest_point_indices() 'x' and 'y' must have shape (N, X, D)")
    N, _, D = x.shape
    if y.shape[0] != N:
        raise ValueError("closest_point_indices() 'x' and 'y' must have same batch size N")
    if y.shape[2] != D:
        raise ValueError("closest_point_indices() 'x' and 'y' must have same point dimension D")
    x = x.float()
    y = y.type(x.dtype)
    indices = torch.empty(x.shape[0:2], dtype=torch.int64, device=x.device)
    for i, points in enumerate(x.split(split_size, dim=1)):
        dists = distance_matrix(points, y)
        j = slice(i * split_size, i * split_size + points.shape[1])
        indices[:, j] = torch.min(dists, dim=2).indices
    return indices


def normalize_grid(
    grid: Tensor,
    size: Optional[Union[Tensor, torch.Size]] = None,
    side_length: float = 2,
    align_corners: bool = ALIGN_CORNERS,
    channels_last: bool = True,
) -> Tensor:
    r"""Map unnormalized grid coordinates to normalized grid coordinates."""
    if not isinstance(grid, Tensor):
        raise TypeError("normalize_grid() 'grid' must be tensors")
    if not grid.is_floating_point():
        grid = grid.float()
    if size is None:
        if channels_last:
            if grid.ndim < 4 or grid.shape[-1] != grid.ndim - 2:
                raise ValueError(
                    "normalize_grid() 'grid' must have shape (N, ..., X, D) when 'size' not given"
                )
            size = torch.Size(reversed(grid.shape[1:-1]))  # X,...
        else:
            if grid.ndim < 4 or grid.shape[1] != grid.ndim - 2:
                raise ValueError(
                    "normalize_grid() 'grid' must have shape (N, D, ..., X) when 'size' not given"
                )
            size = torch.Size(reversed(grid.shape[2:]))  # X,...
    zero = torch.tensor(0, dtype=grid.dtype, device=grid.device)
    size = torch.as_tensor(size, dtype=grid.dtype, device=grid.device)
    size_ = size.sub(1) if align_corners else size
    if not channels_last:
        grid = move_dim(grid, 1, -1)
    if side_length != 1:
        grid = grid.mul(side_length)
    grid = torch.where(size > 1, grid.div(size_).sub(1), zero)
    if not channels_last:
        grid = move_dim(grid, -1, 1)
    return grid


def denormalize_grid(
    grid: Tensor,
    size: Optional[Union[Tensor, torch.Size]] = None,
    side_length: float = 2,
    align_corners: bool = ALIGN_CORNERS,
    channels_last: bool = True,
) -> Tensor:
    r"""Map normalized grid coordinates to unnormalized grid coordinates."""
    if not isinstance(grid, Tensor):
        raise TypeError("denormalize_grid() 'grid' must be tensors")
    if size is None:
        if grid.ndim < 4 or grid.shape[-1] != grid.ndim - 2:
            raise ValueError(
                "normalize_grid() 'grid' must have shape (N, ..., X, D) when 'size' not given"
            )
        size = torch.Size(reversed(grid.shape[1:-1]))  # X,...
    zero = torch.tensor(0, dtype=grid.dtype, device=grid.device)
    size = torch.as_tensor(size, dtype=grid.dtype, device=grid.device)
    size_ = size.sub(1) if align_corners else size
    if not channels_last:
        grid = move_dim(grid, 1, -1)
    grid = torch.where(size > 1, grid.add(1).mul(size_), zero)
    if side_length != 1:
        grid = grid.div(side_length)
    if not channels_last:
        grid = move_dim(grid, -1, 1)
    return grid


def polyline_directions(
    points: Tensor, normalize: bool = False, repeat_last: bool = True
) -> Tensor:
    r"""Compute proximal to distal facing tangent vectors."""
    if not isinstance(points, Tensor):
        raise TypeError("polyline_directions() 'points' must be Tensor")
    if points.ndim < 2:
        raise ValueError("polyline_directions() 'points' must have shape (..., N, 3)")
    dim = points.ndim - 2
    n = points.shape[dim]
    d = points.narrow(dim, 1, n - 1).sub(points.narrow(dim, 0, n - 1))
    if normalize:
        d = F.normalize(d, p=2, dim=dim)
    if repeat_last:
        d = torch.cat([d, d.narrow(dim, n - 2, 1)], dim=dim)
    return d


def polyline_tangents(points: Tensor, normalize: bool = False, repeat_first: bool = True) -> Tensor:
    r"""Compute distal to proximal facing tangent vectors."""
    if not isinstance(points, Tensor):
        raise TypeError("polyline_tangents() 'points' must be Tensor")
    if points.ndim < 2:
        raise ValueError("polyline_tangents() 'points' must have shape (..., N, 3)")
    dim = points.ndim - 2
    n = points.shape[dim]
    d = points.narrow(dim, 0, n - 1).sub(points.narrow(dim, 1, n - 1))
    if normalize:
        d = F.normalize(d, p=2, dim=dim)
    if repeat_first:
        d = torch.cat([d.narrow(dim, 0, 1), d], dim=dim)
    return d


def transform_grid(transform: Tensor, grid: Tensor, align_corners: bool = ALIGN_CORNERS) -> Tensor:
    r"""Transform undeformed grid by a spatial transformation.

    This function applies a spatial transformation to map a tensor of undeformed grid points to a
    tensor of deformed grid points with the same shape as the input tensor. The input points must be
    the positions of undeformed spatial grid points, because in case of a non-rigid transformation,
    this function uses interpolation to resize the vector fields to the size of the input ``grid``.
    This assumes that input points ``x`` are the coordinates of points located on a regularly spaced
    undeformed grid which is aligned with the borders of the grid domain on which the vector fields
    of the non-rigid transformations are sampled, i.e., ``y = x + u``.

    In case of a linear transformation ``y = Ax + t``.

    If in doubt whether the input points will be sampled regularly at grid points of the domain of
    the spatial transformation, use ``transform_points()`` instead.

    Args:
        transform: Tensor representation of spatial transformation, where the shape of the tensor
            determines the type of transformation. A translation-only transformation must be given
            as tensor of shape ``(N, D, 1)``. An affine-only transformation without translation can
            be given as tensor of shape ``(N, D, D)``, and an affine transformation with translation
            as tensor of shape ``(N, D, D + 1)``. Flow fields of non-rigid transformations, on the
            other hand, are tensors of shape ``(N, D, ..., X)``, i.e., linear transformations are
            represented by 3-dimensional tensors, and non-rigid transformations by tensors of at least
            4 dimensions. If batch size is one, but the batch size of ``points`` is greater than one,
            all point sets are transformed by the same non-rigid transformation.
        grid: Coordinates of undeformed grid points as tensor of shape ``(N, ..., D)`` or ``(1, ..., D)``.
            If batch size is one, but multiple flow fields are given, this single point set is
            transformed by each non-rigid transformation to produce ``N`` output point sets.
        align_corners: Whether flow vectors in case of a non-rigid transformation are with respect to
            ``Axes.CUBE`` (False) or ``Axes.CUBE_CORNERS`` (True). The input ``grid`` points must be
            with respect to the same spatial grid domain as the input flow fields. This option is in
            particular passed on to the ``grid_reshape()`` function used to resize the flow fields to
            the shape of the input grid.

    Returns:
        Tensor of shape ``(N, ..., D)`` with coordinates of spatially transformed points.

    """
    if not isinstance(transform, Tensor):
        raise TypeError("transform_grid() 'transform' must be Tensor")
    if transform.ndim < 3:
        raise ValueError("transform_grid() 'transform' must be at least 3-dimensional tensor")
    if transform.ndim == 3:
        return A.transform_points(transform, grid)
    return warp_grid(transform, grid, align_corners=align_corners)


def transform_points(
    transform: Tensor, points: Tensor, align_corners: bool = ALIGN_CORNERS
) -> Tensor:
    r"""Transform set of points by a tensor of non-rigid flow fields.

    This function applies a spatial transformation to map a tensor of points to a tensor of transformed
    points of the same shape as the input tensor. Unlike ``transform_grid()``, it can be used to spatially
    transform any set of points which are defined with respect to the grid domain of the spatial transformation,
    including a tensor of shape ``(N, M, D)``, i.e., a batch of N point sets with cardianality M. It can also
    be applied to a tensor of grid points of shape ``(N, ..., X, D)`` regardless if the grid points are located
    at the undeformed grid positions or an already deformed grid. Therefore, in case of a non-rigid transformation,
    the given flow fields are sampled at the input points ``x`` using linear interpolation. The flow vectors ``u(x)``
    are then added to the input points, i.e., ``y = x + u(x)``.

    In case of a linear transformation ``y = Ax + t``.

    Args:
        transform: Tensor representation of spatial transformation, where the shape of the tensor
            determines the type of transformation. A translation-only transformation must be given
            as tensor of shape ``(N, D, 1)``. An affine-only transformation without translation can
            be given as tensor of shape ``(N, D, D)``, and an affine transformation with translation
            as tensor of shape ``(N, D, D + 1)``. Flow fields of non-rigid transformations, on the
            other hand, are tensors of shape ``(N, D, ..., X)``, i.e., linear transformations are
            represented by 3-dimensional tensors, and non-rigid transformations by tensors of at least
            4 dimensions. If batch size is one, but the batch size of ``points`` is greater than one,
            all point sets are transformed by the same non-rigid transformation.
        points: Coordinates of points given as tensor of shape ``(N, ..., D)`` or ``(1, ..., D)``.
            If batch size is one, but multiple flow fields are given, this single point set is
            transformed by each non-rigid transformation to produce ``N`` output point sets.
        align_corners: Whether flow vectors in case of a non-rigid transformation are with respect to
            ``Axes.CUBE`` (False) or ``Axes.CUBE_CORNERS`` (True). The input ``points`` must be
            with respect to the same spatial grid domain as the input flow fields. This option is in
            particular passed on to the ``grid_sample()`` function used to sample the flow vectors at
            the input points.

    Returns:
        Tensor of shape ``(N, ..., D)`` with coordinates of spatially transformed points.

    """
    if not isinstance(transform, Tensor):
        raise TypeError("transform_points() 'transform' must be Tensor")
    if transform.ndim < 3:
        raise ValueError("transform_points() 'transform' must be at least 3-dimensional tensor")
    if transform.ndim == 3:
        return A.transform_points(transform, points)
    return warp_points(transform, points, align_corners=align_corners)



from typing import NamedTuple, Optional

from boto3.s3.transfer import TransferConfig


KiB = 1024
MiB = 1024**2


class S3Config(NamedTuple):
    r"""Configuration of AWS Simple Storage Service."""

    mode: str = "rwd"
    verify: bool = True
    region: Optional[str] = None
    multipart_threshold: int = 8 * MiB
    max_concurrency: int = 10
    multipart_chunksize: int = 8 * MiB
    num_download_attempts: int = 5
    max_io_queue: int = 100
    io_chunksize: int = 256 * KiB
    use_threads: bool = True
    max_bandwidth: Optional[int] = None

    def transfer_config(self) -> TransferConfig:
        return TransferConfig(
            multipart_threshold=self.multipart_threshold,
            max_concurrency=self.max_concurrency,
            multipart_chunksize=self.multipart_chunksize,
            num_download_attempts=self.num_download_attempts,
            max_io_queue=self.max_io_queue,
            io_chunksize=self.io_chunksize,
            use_threads=self.use_threads,
            max_bandwidth=self.max_bandwidth,
        )



r"""Convoluational modules with fixed parameters."""

import math
from numbers import Number
from typing import Optional, Union

from pkg_resources import parse_version

import torch
from torch import Tensor
from torch.nn import Module, functional as F

from deepali.core import functional as U
from deepali.core.enum import PaddingMode
from deepali.core.kernels import gaussian1d
from deepali.core.typing import ScalarOrTuple


class FilterImage(Module):
    r"""Convoles an image with a predefined filter kernel."""

    def __init__(
        self,
        kernel: Optional[Tensor],
        padding: Optional[Union[PaddingMode, str]] = None,
    ):
        r"""Initialize parameters.

        Args:
            kernel: Predefined convolution kernel.
            padding: Image extrapolation mode.

        """
        super().__init__()
        self.padding = PaddingMode.CONSTANT if padding is None else PaddingMode.from_arg(padding)
        self.register_buffer("kernel", kernel)

    def forward(self, x: Tensor) -> Tensor:
        r"""Convolve input image with predefined filter kernel."""
        kernel: Optional[Tensor] = self.kernel
        if self.kernel is None or kernel.numel() < 2:
            return x
        return U.conv(x, kernel, padding=self.padding)

    def extra_repr(self) -> str:
        return f"padding={repr(self.padding.value)}"


class BlurImage(FilterImage):
    r"""Blurs an image by a predefined Gaussian low-pass filter."""

    def __init__(self, sigma: float, padding: Optional[Union[PaddingMode, str]] = None):
        r"""Initialize parameters.

        Args:
            sigma: Standard deviation of isotropic Gaussian kernel in grid units (pixel, voxel).
            padding: Image extrapolation mode.

        """
        sigma = float(sigma)
        kernel = gaussian1d(sigma) if sigma > 0 else None
        super().__init__(kernel=kernel, padding=padding)
        self.sigma = sigma

    def extra_repr(self) -> str:
        return f"sigma={repr(self.sigma)}, " + super().extra_repr()


class GaussianConv(Module):
    r"""Blurs an image by a predefined Gaussian low-pass filter."""

    def __init__(
        self, channels: int, kernel_size: ScalarOrTuple[int], sigma: float, dim: int = 3
    ) -> None:
        r"""Initialize Gaussian convolution kernel.

        Args:
            channels (int, sequence): Number of channels of the input and output tensors.
            kernel_size (int, sequence): Size of the gaussian kernel.
            sigma (float, sequence): Standard deviation of the gaussian kernel.
            dim (int, optional): The number of dimensions of the data.

        """
        if dim < 2 or dim > 3:
            raise ValueError(f"Only 2 and 3 dimensions are supported, got: {dim}")
        super().__init__()
        if isinstance(kernel_size, Number):
            kernel_size = (kernel_size,) * dim
        if isinstance(sigma, Number):
            sigma = (sigma,) * dim
        # The gaussian kernel is the product of the gaussian function of each dimension.
        kernel = torch.tensor(1, dtype=torch.float32, device="cpu")
        mgrids = [torch.arange(n, dtype=torch.float32) for n in kernel_size]
        if parse_version(torch.__version__) < parse_version("1.10"):
            mgrids = torch.meshgrid(mgrids)
        else:
            mgrids = torch.meshgrid(mgrids, indexing="ij")
        norm = math.sqrt(2 * math.pi)
        for size, std, mgrid in zip(kernel_size, sigma, mgrids):
            mean = (size - 1) / 2
            kernel *= 1 / (std * norm) * torch.exp(-(((mgrid - mean) / std) ** 2) / 2)
        # Make sure sum of values in gaussian kernel equals 1.
        kernel = kernel.div_(kernel.sum())
        # Reshape to depthwise convolutional weight
        kernel = kernel.view(1, 1, *kernel.size())
        kernel = kernel.repeat(channels, *((1,) * (kernel.dim() - 1)))
        self.register_buffer("kernel", kernel, persistent=True)
        self.groups = channels
        # Padding for output to equal input dimensions
        self.pad = (kernel_size[0] // 2,) * (2 * dim)
        self.conv = F.conv2d if dim == 2 else F.conv3d

    def forward(self, x: Tensor) -> Tensor:
        r"""Convolve input with Gaussian kernel."""
        # Use mode='replicate' to ensure image boundaries are not degraded in output for background channel
        kernel: Tensor = self.kernel
        data = F.pad(x, self.pad, mode="replicate")
        data = self.conv(data, kernel, groups=self.groups)
        return data



r"""Hooks for iterative optimization-based registration engine."""

from typing import Callable

import torch.nn.functional as F

from deepali.core import functional as U
from deepali.core.kernels import gaussian1d
from deepali.spatial import is_linear_transform

from .engine import RegistrationEngine, RegistrationResult


RegistrationEvalHook = Callable[[RegistrationEngine, int, int, RegistrationResult], None]
RegistrationStepHook = Callable[[RegistrationEngine, int, int, float], None]


def noop(reg: RegistrationEngine, *args, **kwargs) -> None:
    r"""Dummy no-op loss evaluation hook."""
    ...


def normalize_linear_grad(reg: RegistrationEngine, *args, **kwargs) -> None:
    r"""Loss evaluation hook for normalization of linear transformation gradient after backward pass."""
    denom = None
    for group in reg.optimizer.param_groups:
        for p in (p for p in group["params"] if p.grad is not None):
            max_abs_grad = p.grad.abs().max()
            if denom is None or denom < max_abs_grad:
                denom = max_abs_grad
    if denom is None:
        return
    for group in reg.optimizer.param_groups:
        for p in (p for p in group["params"] if p.grad is not None):
            p.grad /= denom


def normalize_nonrigid_grad(reg: RegistrationEngine, *args, **kwargs) -> None:
    r"""Loss evaluation hook for normalization of non-rigid transformation gradient after backward pass."""
    for group in reg.optimizer.param_groups:
        for p in (p for p in group["params"] if p.grad is not None):
            F.normalize(p.grad, p=2, dim=1, out=p.grad)


def normalize_grad_hook(transform) -> RegistrationEvalHook:
    r"""Loss evaluation hook for normalization of transformation gradient after backward pass."""
    if is_linear_transform(transform):
        return normalize_linear_grad
    return normalize_nonrigid_grad


def _smooth_nonrigid_grad(reg: RegistrationEngine, sigma: float = 1) -> None:
    r"""Loss evaluation hook for Gaussian smoothing of non-rigid transformation gradient after backward pass."""
    if sigma <= 0:
        return
    kernel = gaussian1d(sigma)
    for group in reg.optimizer.param_groups:
        for p in (p for p in group["params"] if p.grad is not None):
            p.grad.copy_(U.conv(p.grad, kernel))


def smooth_grad_hook(transform, sigma: float) -> RegistrationEvalHook:
    r"""Loss evaluation hook for Gaussian smoothing of non-rigid gradient after backward pass."""
    if is_linear_transform(transform):
        return noop

    def fn(reg: RegistrationEngine, *args, **kwargs):
        return _smooth_nonrigid_grad(reg, sigma=sigma)

    return fn



r"""Image tensors."""

from __future__ import annotations

from typing import Any, Dict, Generator, Optional, Sequence, Tuple, Type, TypeVar, Union, overload

import numpy as np
import torch
from torch import Tensor
from torch.nn import functional as F

from deepali.core.cube import Cube
from deepali.core.enum import PaddingMode, Sampling, SpatialDimArg
from deepali.core.grid import ALIGN_CORNERS, Axes, Grid, grid_transform_points
from deepali.core import image as U
from deepali.core.itertools import zip_longest_repeat_last
from deepali.core.tensor import cat_scalars
from deepali.core.typing import Array, Device, DType, EllipsisType
from deepali.core.typing import PathUri, Scalar, ScalarOrTuple, Size
from deepali.utils.imageio import read_image, write_image

from .tensor import DataTensor

Domain = Cube

TImage = TypeVar("TImage", bound="Image")
TImageBatch = TypeVar("TImageBatch", bound="ImageBatch")


__all__ = ("Image", "ImageBatch")


class ImageBatch(DataTensor):
    r"""Batch of images sampled on regular oriented grids."""

    def __init__(
        self: TImageBatch,
        data: Array,
        grid: Optional[Union[Grid, Sequence[Grid]]] = None,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
        requires_grad: Optional[bool] = None,
        pin_memory: bool = False,
    ) -> None:
        r"""Initialize image decorator.

        Args:
            data: Image batch data tensor of shape (N, C, ...X).
            grid: Sampling grid of image data oriented in world space. Can be either a single shared
                sampling grid, or a separate grid for each image in the batch. Note that operations
                which would result in differently sized images (e.g., resampling to a certain voxel
                size, when images have different resolutions) will raise an exception. All images in
                a batch must have the same number of channels and spatial size. If ``None``, a default
                grid whose world space coordinate axes are aligned with the image axes, unit spacing,
                and origin at the image centers is created. By default, image grid attributes are always
                stored in CPU memory, regardless of the ``device`` on which the image data is located.
            dtype: Data type of the image data. A copy of the data is only made when the desired ``dtype``
                is not ``None`` and not the same as ``data.dtype``.
            device: Device on which to store image data. A copy of the data is only made when the data
                has to be copied to a different device.
            requires_grad: If autograd should record operations on the returned image tensor.
            pin_memory: If set, returned image tensor would be allocated in the pinned memory.
                Works only for CPU tensors.

        """
        # DataTensor.__new__() creates the tensor subclass given arguments:
        # data, dtype, device, requires_grad, pin_memory
        if self.ndim < 4:
            raise ValueError("Image batch tensor must have at least four dimensions")
        self.grid_(grid)

    def _make_instance(
        self: TImageBatch,
        data: Optional[Tensor] = None,
        grid: Optional[Sequence[Grid]] = None,
        **kwargs,
    ) -> TImageBatch:
        r"""Create a new instance while preserving subclass (meta-)data."""
        kwargs["grid"] = self._grid if grid is None else grid
        return super()._make_instance(data, **kwargs)

    def _make_subitem(self, data: Tensor, grid: Grid) -> Image:
        r"""Create Image in __getitem__. Can be overridden by subclasses to return a subtype."""
        return Image(data, grid)

    def __deepcopy__(self: TImageBatch, memo) -> TImageBatch:
        if id(self) in memo:
            return memo[id(self)]
        result = self._make_instance(
            self.data.clone(memory_format=torch.preserve_format),
            grid=tuple(grid.clone() for grid in self._grid),
            requires_grad=self.requires_grad,
            pin_memory=self.is_pinned(),
        )
        memo[id(self)] = result
        return result

    @staticmethod
    def _torch_function_grid(
        func, args, kwargs: Dict[str, Any]
    ) -> Optional[Union[Sequence[Grid], Sequence[Sequence[Grid]]]]:
        r"""Get spatial sampling grids from args passed to __torch_function__."""
        if not args:
            return None
        if isinstance(args[0], (tuple, list)):
            args = args[0]
        grids: Sequence[Sequence[Grid]]
        grids = [g for g in (getattr(arg, "_grid", None) for arg in args) if g is not None]
        if not grids:
            return None
        if kwargs.get("dim", 0) == 0:
            if func == torch.cat:
                return [g for grid in grids for g in grid]
            if func in (torch.split, Tensor.split):
                grids = grids[0]
                split_grids = []
                split_size_or_sections = args[1]
                if isinstance(split_size_or_sections, int):
                    for start in range(0, len(grids), split_size_or_sections):
                        split_grids.append(grids[start : start + split_size_or_sections])
                elif isinstance(split_size_or_sections, Sequence):
                    start = 0
                    for num in split_size_or_sections:
                        split_grids.append(grids[start : start + num])
                return split_grids
            if func in (torch.split_with_sizes, Tensor.split_with_sizes):
                grids = grids[0]
                split_grids = []
                split_sizes = args[1]
                start = 0
                for num in split_sizes:
                    split_grids.append(grids[start : start + num])
                return split_grids
            if func in (torch.tensor_split, Tensor.tensor_split):
                grids = grids[0]
                split_grids = []
                tensor_indices_or_sections = args[1]
                if isinstance(tensor_indices_or_sections, int):
                    for start in range(0, len(grids), tensor_indices_or_sections):
                        split_grids.append(grids[start : start + tensor_indices_or_sections])
                elif isinstance(tensor_indices_or_sections, Sequence):
                    indices = list(tensor_indices_or_sections)
                    for start, end in zip([0] + indices, indices + [len(grids)]):
                        split_grids.append(grids[start:end])
                return split_grids
        return grids[0]

    @classmethod
    def _torch_function_result(cls, func, data, grid: Optional[Sequence[Grid]]) -> Any:
        if not isinstance(data, Tensor):
            return data
        if (
            grid
            and data.ndim == grid[0].ndim + 2
            and data.shape[0] == len(grid)
            and data.shape[2:] == grid[0].shape
        ) or (grid is not None and len(grid) == 0 and data.ndim >= 4 and data.shape[0] == 0):
            if func in (torch.clone, Tensor.clone):
                grid = [g.clone() for g in grid]
            if isinstance(data, cls):
                data._grid = grid
            else:
                data = cls(data, grid)
        elif type(data) != Tensor:
            data = data.as_subclass(Tensor)
        return data

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        args = tuple(arg.batch() if isinstance(arg, Image) else arg for arg in args)
        # fargs = [arg.as_subclass(Tensor) if isinstance(arg, Tensor) else arg for arg in args]
        data = Tensor.__torch_function__(func, (Tensor,), args, kwargs)
        grid = cls._torch_function_grid(func, args, kwargs)
        if func in (F.grid_sample,):
            grid = None
        elif func in (
            torch.split,
            Tensor.split,
            torch.split_with_sizes,
            Tensor.split_with_sizes,
            torch.tensor_split,
            Tensor.tensor_split,
        ):
            if type(data) not in (tuple, list):
                raise AssertionError(f"expected split 'data' to be tuple or list, got {type(data)}")
            if type(grid) not in (tuple, list):
                raise AssertionError(f"expected split 'grid' to be tuple or list, got {type(grid)}")
            if len(grid) != len(data):
                raise AssertionError(
                    f"expected 'grid' tuple length to be equal batch size, but {len(grid)} != {len(data)}"
                )
            assert all(isinstance(d, Tensor) for d in data)
            assert all(isinstance(g, (tuple, list)) for g in grid)
            assert all(len(d) == len(g) for d, g in zip(data, grid))
            return tuple(cls._torch_function_result(func, d, g) for d, g in zip(data, grid))
        return cls._torch_function_result(func, data, grid)

    @classmethod
    def from_images(cls: Type[TImageBatch], images: Sequence[Image]) -> TImageBatch:
        r"""Create image batch from sequence of images."""
        if not all(isinstance(image, DataTensor) for image in images):
            raise TypeError(f"{cls.__name__}.from_images() 'images' must be Image sequence")
        data = torch.cat([image.tensor().unsqueeze(0) for image in images], dim=0)
        grid = [image.grid() for image in images]
        return cls(data, grid)

    def append(self: TImageBatch, other: ImageBatch) -> TImageBatch:
        r"""Append data from another image batch to data of this batch."""
        if not isinstance(other, ImageBatch):
            raise TypeError(f"{type(self).__name__}.append() 'other' must be ImageBatch")
        data = torch.cat([self.tensor(), other.tensor()], dim=0)
        grid = self.grids() + other.grids()
        return self._make_instance(data, grid)

    def cube(self: TImageBatch, n: int = 0) -> Cube:
        r"""Get cube of n-th image in batch defining its normalized coordinates space with respect to the world."""
        return self._grid[n].cube()

    def cubes(self: TImageBatch) -> Tuple[Cube, ...]:
        r"""Get cubes of all images which define their normalized coordinates spaces with respect to the world."""
        return tuple(grid.cube() for grid in self._grid)

    def domain(self: TImageBatch, n: int = 0) -> Domain:
        r"""Get oriented bounding box of n-th image in world space which defines the domain within which it is defined."""
        return self._grid[n].domain()

    def domains(self: TImageBatch) -> Tuple[Domain, ...]:
        r"""Get oriented bounding boxes of all images in world space which define the domains within which these are defined."""
        return tuple(grid.domain() for grid in self._grid)

    def same_domains_as(self, other: ImageBatch) -> bool:
        """Check if images in this batch and another batch have the same cube domains."""
        if len(self) != len(other):
            return False
        return all(a.same_domain_as(b) for a, b in zip(self.grids(), other.grids()))

    @overload
    def grid(self: TImageBatch, n: int = 0) -> Grid:
        r"""Get sampling grid of n-th image in batch."""
        ...

    @overload
    def grid(self: TImageBatch, arg: Union[Grid, Sequence[Grid]]) -> TImageBatch:
        r"""Get new image batch with specified sampling grid(s)."""
        ...

    def grid(
        self: TImageBatch, arg: Optional[Union[int, Grid, Sequence[Grid]]] = None
    ) -> Union[Grid, TImageBatch]:
        r"""Get sampling grid of images in batch or new batch with specified grid, respectively."""
        if arg is None:
            arg = 0
        if isinstance(arg, int):
            return self._grid[arg]
        return self._make_instance(grid=arg)

    def grid_(self: TImageBatch, arg: Union[Grid, Sequence[Grid], None]) -> TImageBatch:
        r"""Change image sampling grid of this image batch."""
        shape = self.shape
        if arg is None:
            arg = (Grid(shape=shape[2:]),) * shape[0]
        elif isinstance(arg, Grid):
            grid = arg
            if grid.shape != shape[2:]:
                raise ValueError(
                    "Image grid size does not match spatial dimensions of image batch tensor"
                )
            arg = (grid,) * shape[0]
        else:
            arg = tuple(arg)
            if any(grid.shape != shape[2:] for grid in arg):
                raise ValueError(
                    "Image grid sizes must match spatial dimensions of image batch tensor"
                )
        self._grid = arg
        return self

    def grids(self: TImageBatch) -> Tuple[Grid, ...]:
        r"""Get sampling grids of images in batch."""
        return self._grid

    def align_corners(self: TImageBatch) -> bool:
        r"""Whether image resizing operations by default preserve corner points or grid extent."""
        return self._grid[0].align_corners()

    def center(self: TImageBatch) -> Tensor:
        r"""Image centers in world space coordinates as tensor of shape (N, D)."""
        return torch.cat([grid.center().unsqueeze(0) for grid in self.grids()], dim=0)

    def origin(self: TImageBatch) -> Tensor:
        r"""Image origins in world space coordinates as tensor of shape (N, D)."""
        return torch.cat([grid.origin().unsqueeze(0) for grid in self.grids()], dim=0)

    def spacing(self: TImageBatch) -> Tensor:
        r"""Image spacings in world units as tensor of shape (N, D)."""
        return torch.cat([grid.spacing().unsqueeze(0) for grid in self.grids()], dim=0)

    def direction(self: TImageBatch) -> Tensor:
        r"""Image direction cosines matrices as tensor of shape (N, D, D)."""
        return torch.cat([grid.direction().unsqueeze(0) for grid in self.grids()], dim=0)

    def __len__(self: TImageBatch) -> int:
        r"""Number of images in batch."""
        return self.shape[0]

    @overload
    def __getitem__(self: TImageBatch, index: int) -> Image:
        ...

    @overload
    def __getitem__(self: TImageBatch, index: EllipsisType) -> TImageBatch:
        ...

    @overload
    def __getitem__(self: TImageBatch, index: Union[list, slice, Tensor]) -> TImageBatch:
        ...

    def __getitem__(
        self: TImageBatch,
        index: Union[EllipsisType, int, slice, Sequence[Union[EllipsisType, int, slice]]],
    ) -> Union[Image, TImageBatch, Tensor]:
        r"""Get image at specified batch index, get a sub-batch, or a region of interest tensor."""
        if index is ...:
            return self._make_instance(self.tensor(), self.grid())
        if type(index) is tuple:
            # Resolve additional ellipses
            index = [j for i, j in enumerate(index) if j is not ... or ... not in index[:i]]
            # Discard trailing ellipsis
            if index[-1] is ...:
                index = index[:-1]
            # Substitute remaining ellipsis with full slices
            try:
                i = index.index(...)
                j = len(index) - i - 1
                index = index[:i] + [slice(None)] * (self.ndim - i - j) + index[-j:]
            except ValueError:
                pass
            index = tuple(index)
            is_multi_index = True
        elif isinstance(index, (np.ndarray, slice, Sequence, Tensor)):
            # - batch[[0, 2, 4]]
            # - batch[np.array([0, 2, 4])]
            # - batch[torch.tensor([0, 2, 4])]
            index = (index,)
            is_multi_index = True
        else:
            index = int(index)
            is_multi_index = False
        data = self.tensor()[index]
        if is_multi_index and len(index) > 1 and isinstance(index[1], int):
            return data  # cannot be an ImageBatch or Image without a channel dimension
        grid_index = index[0] if is_multi_index else index
        if isinstance(grid_index, (np.ndarray, Sequence, Tensor)):
            grid = tuple(self._grid[i] for i in grid_index)
        else:
            grid = self._grid[grid_index]
        if is_multi_index and len(index) > 2:
            same_grid = True
            for i, n in zip(index[2:], self.shape[2:]):
                if isinstance(i, int) or not isinstance(i, slice):
                    same_grid = False
                    break
                if i.start not in (None, 0) or i.stop not in (None, n) or i.step not in (None, 1):
                    same_grid = False
                    break
            if not same_grid:
                return data
        if isinstance(grid, Grid):
            if data.ndim < 3:
                return data
            return self._make_subitem(data, grid)
        elif data.ndim < 4:
            return data
        return self._make_instance(data, grid)

    def __iter__(self) -> Generator[Image, None, None]:
        r"""Generator to iterate over images in batch."""
        for index in range(len(self)):
            data = self.tensor().narrow(0, index, 1).squeeze_(0)
            yield self._make_subitem(data, self._grid[index])

    @property
    def sdim(self: TImageBatch) -> int:
        r"""Number of spatial dimensions."""
        return self.ndim - 2

    @property
    def nchannels(self: TImageBatch) -> int:
        r"""Number of image channels."""
        return self.shape[1]

    def normalize(
        self: TImageBatch,
        mode: str = "unit",
        min: Optional[float] = None,
        max: Optional[float] = None,
    ) -> TImageBatch:
        r"""Normalize image intensities in [min, max]."""
        return U.normalize_image(self, mode=mode, min=min, max=max)

    def normalize_(
        self: TImageBatch,
        mode: str = "unit",
        min: Optional[float] = None,
        max: Optional[float] = None,
    ) -> TImageBatch:
        r"""Normalize image intensities in [min, max]."""
        return U.normalize_image(self, mode=mode, min=min, max=max, inplace=True)

    def rescale(
        self: TImageBatch,
        min: Optional[Scalar] = None,
        max: Optional[Scalar] = None,
        data_min: Optional[Scalar] = None,
        data_max: Optional[Scalar] = None,
        dtype: Optional[DType] = None,
    ) -> TImageBatch:
        r"""Clamp and linearly rescale image values."""
        return U.rescale(self, min, max, data_min=data_min, data_max=data_max, dtype=dtype)

    def narrow(self: TImageBatch, dim: int, start: int, length: int) -> TImageBatch:
        r"""Narrow image batch along specified tensor dimension."""
        data = self.tensor().narrow(dim, start, length)
        grid = self.grid()
        if dim > 1:
            grid = grid.narrow(self.ndim - dim - 1, start, length)
        return self._make_instance(data, grid)

    def resize(
        self: TImageBatch,
        size: Union[int, Array, Size],
        *args: int,
        mode: Union[Sampling, str] = Sampling.LINEAR,
        align_corners: Optional[bool] = None,
    ) -> TImageBatch:
        r"""Interpolate images on grid with specified size.

        Args:
            size: Size of spatial image dimensions, where the size of the last tensor dimension,
                which corresponds to the first grid dimension, must be given first, e.g., ``(nx, ny, nz)``.
            mode: Image data interpolation mode.
            align_corners: Whether to preserve grid extent (False) or corner points (True).
                If ``None``, the default of the image sampling grid is used.

        Returns:
            Image batch with specified size of spatial dimensions.

        """
        if align_corners is None:
            align_corners = self.align_corners()
        size = cat_scalars(size, *args, num=self.sdim, device=self.device)
        data = U.grid_resize(self, size, mode=mode, align_corners=align_corners)
        grid = tuple(grid.resize(size, align_corners=align_corners) for grid in self._grid)
        return self._make_instance(data, grid)

    def resample(
        self: TImageBatch,
        spacing: Union[float, Array, str],
        *args: float,
        mode: Union[Sampling, str] = Sampling.LINEAR,
    ) -> TImageBatch:
        r"""Interpolate images on grid with specified spacing.

        Args:
            spacing: Spacing of grid on which to resample image data, where the spacing corresponding
                to first grid dimension, which corresponds to the last tensor dimension, must be given
                first, e.g., ``(sx, sy, sz)``. Alternatively, can be string 'min' or 'max' to resample
                to the minimum or maximum voxel size, respectively.
            mode: Image data interpolation mode.

        Returns:
            Image batch with given grid spacing.

        """
        in_spacing = self._grid[0].spacing()
        if not all(torch.allclose(grid.spacing(), in_spacing) for grid in self._grid):
            raise AssertionError(
                f"{type(self).__name__}.resample() requires all images in batch to have the same grid spacing"
            )
        if spacing == "min":
            assert not args
            out_spacing = in_spacing.min()
        elif spacing == "max":
            assert not args
            out_spacing = in_spacing.max()
        else:
            out_spacing = spacing
        out_spacing = cat_scalars(out_spacing, *args, num=self.sdim, device=self.device)
        data = U.grid_resample(self, in_spacing=in_spacing, out_spacing=out_spacing, mode=mode)
        grid = tuple(grid.resample(out_spacing) for grid in self._grid)
        return self._make_instance(data, grid)

    def avg_pool(
        self: TImageBatch,
        kernel_size: ScalarOrTuple[int],
        stride: Optional[ScalarOrTuple[int]] = None,
        padding: ScalarOrTuple[int] = 0,
        ceil_mode: bool = False,
        count_include_pad: bool = True,
    ) -> TImageBatch:
        r"""Average pooling of image data."""
        data = U.avg_pool(
            self,
            kernel_size,
            stride=stride,
            padding=padding,
            ceil_mode=ceil_mode,
            count_include_pad=count_include_pad,
        )
        grid = tuple(
            grid.avg_pool(
                kernel_size,
                stride=stride,
                padding=padding,
                ceil_mode=ceil_mode,
            )
            for grid in self._grid
        )
        return self._make_instance(data, grid)

    def downsample(
        self: TImageBatch,
        levels: int = 1,
        dims: Optional[Sequence[SpatialDimArg]] = None,
        sigma: Optional[Union[Scalar, Array]] = None,
        mode: Optional[Union[Sampling, str]] = None,
        min_size: int = 0,
        align_corners: Optional[bool] = None,
    ) -> TImageBatch:
        r"""Downsample images in batch by halving their size the specified number of times.

        Args:
            levels: Number of times the image size is halved (>0) or doubled (<0).
            dims: Spatial dimensions along which to downsample. If not specified, consider all spatial dimensions.
            sigma: Standard deviation of Gaussian filter applied at each downsampling level.
            mode: Image interpolation mode.
            align_corners: Whether to preserve grid extent (False) or corner points (True).
                If ``None``, the default of the image sampling grid is used.

        Returns:
            Batch of downsampled images.

        """
        if not isinstance(levels, int):
            raise TypeError(f"{type(self).__name__}.downsample() 'levels' must be of type int")
        if align_corners is None:
            align_corners = self.align_corners()
        data = U.downsample(
            self,
            levels,
            dims=dims,
            sigma=sigma,
            mode=mode,
            min_size=min_size,
            align_corners=align_corners,
        )
        grid = tuple(
            grid.downsample(levels, dims=dims, min_size=min_size, align_corners=align_corners)
            for grid in self._grid
        )
        return self._make_instance(data, grid)

    def upsample(
        self: TImageBatch,
        levels: int = 1,
        dims: Optional[Sequence[SpatialDimArg]] = None,
        sigma: Optional[Union[Scalar, Array]] = None,
        mode: Optional[Union[Sampling, str]] = None,
        align_corners: Optional[bool] = None,
    ) -> TImageBatch:
        r"""Upsample image in batch by doubling their size the specified number of times.

        Args:
            levels: Number of times the image size is doubled (>0) or halved (<0).
            dims: Spatial dimensions along which to upsample. If not specified, consider all spatial dimensions.
            sigma: Standard deviation of Gaussian filter applied at each downsampling level.
            mode: Image interpolation mode.
            align_corners: Whether to preserve grid extent (False) or corner points (True).
                If ``None``, the default of the image sampling grid is used.

        Returns:
            Batch of upsampled images.

        """
        if not isinstance(levels, int):
            raise TypeError(f"{type(self).__name__}.upsample() 'levels' must be of type int")
        if align_corners is None:
            align_corners = self.align_corners()
        data = U.upsample(
            self, levels, dims=dims, sigma=sigma, mode=mode, align_corners=align_corners
        )
        grid = tuple(
            grid.upsample(levels, dims=dims, align_corners=align_corners) for grid in self._grid
        )
        return self._make_instance(data, grid)

    def pyramid(
        self: TImageBatch,
        levels: int,
        start: int = 0,
        end: int = -1,
        dims: Optional[Sequence[SpatialDimArg]] = None,
        sigma: Optional[Union[Scalar, Array]] = None,
        mode: Optional[Union[Sampling, str]] = None,
        spacing: Optional[float] = None,
        min_size: int = 0,
        align_corners: Optional[bool] = None,
    ) -> Dict[int, TImageBatch]:
        r"""Create Gaussian resolution pyramid.

        Args:
            levels: Number of resolution levels.
            start: Highest resolution level to return, where 0 corresponds to the finest resolution.
            end: Lowest resolution level to return (inclusive).
            dims: Spatial dimensions along which to downsample. If not specified, consider all spatial dimensions.
            sigma: Standard deviation of Gaussian filter applied at each downsampling level.
            mode: Interpolation mode for resampling image data on downsampled grid.
            spacing: Grid spacing at finest resolution level. Note that this option may increase the
                cube extent of the multi-resolution pyramid sampling grids.
            min_size: Minimum grid size.
            align_corners: Whether to preserve grid extent (False) or corner points (True).
                If ``None``, the default of the image sampling grid is used.

        Returns:
            Dictionary of downsampled image batches with keys corresponding to level indices.

        """
        if not isinstance(levels, int):
            raise TypeError(f"{type(self).__name__}.pyramid() 'levels' must be int")
        if not isinstance(start, int):
            raise TypeError(f"{type(self).__name__}.pyramid() 'start' must be int")
        if not isinstance(end, int):
            raise TypeError(f"{type(self).__name__}.pyramid() 'end' must be int")
        if start < 0:
            start = levels + start
        if start < 0 or start > levels - 1:
            raise ValueError(
                f"{type(self).__name__}.pyramid() 'start' must be in [{-levels}, {levels - 1}]"
            )
        if end < 0:
            end = levels + end
        if end < 0 or end > levels - 1:
            raise ValueError(
                f"{type(self).__name__}.pyramid() 'end' must be in [{-levels}, {levels - 1}]"
            )
        if start > end:
            return {}
        # Current image grids
        if align_corners is None:
            align_corners = self.align_corners()
        grids = tuple(grid.align_corners(align_corners) for grid in self._grid)
        # Finest level grids of multi-level resolution pyramid
        if spacing is not None:
            spacing0 = grids[0].spacing()
            if not all(torch.allclose(grid.spacing(), spacing0) for grid in grids):
                raise AssertionError(
                    f"{type(self).__name__}.pyramid() requires all images to have the same grid"
                    " spacing when output 'spacing' at finest level is specified"
                )
            grids = tuple(grid.resample(spacing) for grid in grids)
        grids = tuple(grid.pyramid(levels, dims=dims, min_size=min_size)[0] for grid in grids)
        assert all(grid.size() == grids[0].size() for grid in grids)
        # Resize image to match finest resolution grid
        if torch.allclose(grids[0].cube_extent(), self._grid[0].cube_extent()):
            size = grids[0].size()
            data = U.grid_resize(self, size, mode=mode, align_corners=align_corners)
        else:
            points = grids[0].coords(device=self.device)
            data = U.grid_sample(self, points, mode=mode, align_corners=align_corners)
        # Construct image pyramid by repeated downsampling
        pyramid = {}
        batch = self._make_instance(data, grids)
        if start == 0:
            pyramid[0] = batch
        for level in range(1, end + 1):
            batch = batch.downsample(dims=dims, sigma=sigma, mode=mode, min_size=min_size)
            if level >= start:
                pyramid[level] = batch
        return pyramid

    def crop(
        self: TImageBatch,
        margin: Optional[Union[int, Array]] = None,
        num: Optional[Union[int, Array]] = None,
        mode: Union[PaddingMode, str] = PaddingMode.CONSTANT,
        value: Scalar = 0,
    ) -> TImageBatch:
        r"""Crop images at boundary.

        Args:
            margin: Number of spatial grid points to remove (positive) or add (negative) at each border.
                Use instead of ``num`` in order to symmetrically crop the input ``data`` tensor, e.g.,
                ``(nx, ny, nz)`` is equivalent to ``num=(nx, nx, ny, ny, nz, nz)``.
            num: Number of spatial gird points to remove (positive) or add (negative) at each border,
                where margin of the last dimension of the ``data`` tensor must be given first, e.g.,
                ``(nx, nx, ny, ny)``. If a scalar is given, the input is cropped equally at all borders.
                Otherwise, the given sequence must have an even length.
            mode: Image extrapolation mode in case of negative crop value.
            value: Constant value used for extrapolation if ``mode=PaddingMode.CONSTANT``.

        Returns:
            Batch of images with modified size, but unchanged spacing.

        """
        data = U.crop(self, margin=margin, num=num, mode=mode, value=value)
        grid = tuple(grid.crop(margin=margin, num=num) for grid in self._grid)
        return self._make_instance(data, grid)

    def pad(
        self: TImageBatch,
        margin: Optional[Union[int, Array]] = None,
        num: Optional[Union[int, Array]] = None,
        mode: Union[PaddingMode, str] = PaddingMode.CONSTANT,
        value: Scalar = 0,
    ) -> TImageBatch:
        r"""Pad images at boundary.

        Args:
            margin: Number of spatial grid points to add (positive) or remove (negative) at each border,
                Use instead of ``num`` in order to symmetrically pad the input ``data`` tensor.
            num: Number of spatial gird points to add (positive) or remove (negative) at each border,
                where margin of the last dimension of the ``data`` tensor must be given first, e.g.,
                ``(nx, ny, nz)``. If a scalar is given, the input is padded equally at all borders.
                Otherwise, the given sequence must have an even length.
            mode: Image extrapolation mode in case of positive pad value.
            value: Constant value used for extrapolation if ``mode=PaddingMode.CONSTANT``.

        Returns:
            Batch of images with modified size, but unchanged spacing.

        """
        data = U.pad(self, margin=margin, num=num, mode=mode, value=value)
        grid = tuple(grid.pad(margin=margin, num=num) for grid in self._grid)
        return self._make_instance(data, grid)

    def center_crop(self: TImageBatch, size: Union[int, Array], *args: int) -> TImageBatch:
        r"""Crop image to specified maximum size.

        Args:
            size: Maximum output size, where the size of the last tensor
                dimension must be given first, i.e., ``(X, ...)``.
                If an ``int`` is given, all spatial output dimensions
                are cropped to this maximum size. If the length of size
                is less than the spatial dimensions of the ``data`` tensor,
                then only the last ``len(size)`` dimensions are modified.

        Returns:
            Batch of cropped images.

        """
        size = cat_scalars(size, *args, num=self.sdim, device=self.device)
        data = U.center_crop(self.tensor(), size)
        grid = tuple(grid.center_crop(size) for grid in self._grid)
        return self._make_instance(data, grid)

    def center_pad(
        self: TImageBatch,
        size: Union[int, Array],
        *args: int,
        mode: Union[PaddingMode, str] = PaddingMode.CONSTANT,
        value: Scalar = 0,
    ) -> TImageBatch:
        r"""Pad image to specified minimum size.

        Args:
            size: Minimum output size, where the size of the last tensor
                dimension must be given first, i.e., ``(X, ...)``.
                If an ``int`` is given, all spatial output dimensions
                are cropped to this maximum size. If the length of size
                is less than the spatial dimensions of the ``data`` tensor,
                then only the last ``len(size)`` dimensions are modified.
            mode: Padding mode (cf. ``torch.nn.functional.pad()``).
            value: Value for padding mode "constant".

        Returns:
            Batch of images with modified size, but unchanged spacing.

        """
        size = cat_scalars(size, *args, num=self.sdim, device=self.device)
        data = U.center_pad(self, size, mode=mode, value=value)
        grid = tuple(grid.center_pad(size) for grid in self._grid)
        return self._make_instance(data, grid)

    def region_of_interest(
        self: TImageBatch,
        start: Union[int, Array],
        size: Union[int, Array],
        padding: Union[PaddingMode, str, float] = PaddingMode.CONSTANT,
        value: float = 0,
    ) -> TImageBatch:
        r"""Extract region of interest from images in batch.

        Args:
            start: Indices of first spatial point to include in region of interest.
            size: Size of region of interest.
            padding: Image extrapolation mode or fill value.
            value: Fill value to use when ``padding=Padding.CONSTANT``.

        Returns:
            Image batch of extracted image region of interest.

        """
        data = U.region_of_interest(self, start, size, padding=padding, value=value)
        grid = tuple(grid.region_of_interest(start, size) for grid in self._grid)
        return self._make_instance(data, grid)

    def conv(
        self: TImageBatch,
        kernel: Union[Tensor, Sequence[Optional[Tensor]]],
        padding: Union[PaddingMode, str, int] = None,
    ) -> TImageBatch:
        r"""Filter images in batch with a given convolutional kernel.

        Args:
            kernel: Weights of kernel used to filter the images in this batch by.
                The dtype of the kernel defines the intermediate data type used for convolutions.
                If a 1-dimensional kernel is given, it is used as seperable convolution kernel in
                all spatial image dimensions. Otherwise, the kernel is applied to the last spatial
                image dimensions. For example, a 2D kernel applied to a batch of 3D image volumes
                is applied slice-by-slice by convolving along the X and Y image axes.
            padding: Image padding mode or margin size. If ``None``, use default mode ``PaddingMode.ZEROS``.

        Returns:
            Result of filtering operation with data type set to the image data type before convolution.
            If this data type is not a floating point data type, the filtered data is rounded and clamped
            before it is being cast to the original dtype.

        """
        data = U.conv(self, kernel, padding=padding)
        crop = tuple((m - n) // 2 for m, n in zip(self.shape[2:], data.shape[2:]))
        crop = tuple(reversed(crop))
        grid = tuple(grid.crop(crop) for grid in self._grid)
        return self._make_instance(data, grid)

    @overload
    def sample(
        self: TImageBatch,
        grid: Union[Grid, Sequence[Grid]],
        mode: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str, Scalar]] = None,
    ) -> TImageBatch:
        r"""Sample images at optionally deformed unit grid points.

        Args:
            grid: Spatial grid points at which to sample image values.
            mode: Image interpolation mode.
            padding: Image extrapolation mode or scalar padding value.

        Returns:
            A new image batch of the resampled data with the given sampling grids.

        """
        ...

    @overload
    def sample(
        self: TImageBatch,
        coords: Tensor,
        mode: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str, Scalar]] = None,
    ) -> Tensor:
        r"""Sample images at optionally deformed unit grid points.

        Args:
            coords: Normalized coordinates at which to sample image values as tensor of
                shape ``(N, ..., D)`` or ``(1, ..., D)``. Note that the shape ``...`` need
                not correspond to a (deformed) grid as required by ``grid_sample()``.
                It can be an arbitrary shape, e.g., ``M`` to sample at ``M`` given points.
            mode: Image interpolation mode.
            padding: Image extrapolation mode or scalar padding value.

        Returns:
            A tensor of shape (N, C, ...) of sampled image values.

        """
        ...

    def sample(
        self: TImageBatch,
        arg: Union[Grid, Sequence[Grid], Tensor],
        mode: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str, Scalar]] = None,
    ) -> Union[TImageBatch, Tensor]:
        r"""Sample images at optionally deformed unit grid points.

        Args:
            arg: Either a single grid which defines the sampling points for all images in the batch,
                a different grid for each image in the batch, or a tensor of normalized coordinates
                with shape ``(N, ..., D)`` or ``(1, ..., D)``. In the latter case, note that the
                shape ``...`` need not correspond to a (deformed) grid as required by ``grid_sample()``.
                It can be an arbitrary shape, e.g., ``M`` to sample at ``M`` given points.
            mode: Image interpolation mode.
            padding: Image extrapolation mode or scalar padding value.

        Returns:
            If ``arg`` is of type ``Grid`` or ``Sequence[Grid]``, an ``ImageBatch`` is returned.
            When these grids match the grids of this image batch, ``self`` is returned.
            Otherwise, a ``Tensor`` of shape (N, C, ...) of sampled image values is returned.

        """
        data = self.tensor()
        align_corners = self.align_corners()
        if isinstance(arg, Tensor):
            return U.sample_image(
                data, arg, mode=mode, padding=padding, align_corners=align_corners
            )
        if isinstance(arg, Grid):
            arg = (arg,)
        elif not isinstance(arg, Sequence) or any(not isinstance(item, Grid) for item in arg):
            raise TypeError(
                f"{type(self).__name__}.sample() 'arg' must be Grid, Sequence[Grid], or Tensor"
            )
        if len(arg) not in (1, len(self)):
            raise ValueError(
                f"{type(self).__name__}.sample() 'arg' must be one or {len(self)} grids"
            )
        if all(grid == g for grid, g in zip_longest_repeat_last(arg, self._grid)):
            return self
        axes = Axes.from_align_corners(align_corners)
        coords = [grid.coords(align_corners=align_corners, device=self.device) for grid in arg]
        coords = torch.cat(
            [
                grid_transform_points(p, grid, axes, to_grid, axes).unsqueeze(0)
                for p, grid, to_grid in zip_longest_repeat_last(coords, arg, self._grid)
            ],
            dim=0,
        )
        data = U.grid_sample(
            data,
            coords,
            mode=mode,
            padding=padding,
            align_corners=align_corners,
        )
        return self._make_instance(data, arg)

    def __repr__(self) -> str:
        return type(self).__name__ + f"(data={self.tensor()!r}, grid={self.grids()!r})"

    def __str__(self) -> str:
        return type(self).__name__ + f"(data={self.tensor()!s}, grid={self.grids()!s})"


class Image(DataTensor):
    r"""Image sampled on oriented grid."""

    def __init__(
        self: TImage,
        data: Array,
        grid: Optional[Grid] = None,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
        requires_grad: Optional[bool] = None,
        pin_memory: bool = False,
    ) -> None:
        r"""Initialize image tensor.

        Args:
            data: Image data tensor of shape (C, ...X). To create an ``Image`` instance from an
                image of a mini-batch without creating a copy of the data, simply provide the
                respective slice of the mini-batch corresponding to this image, e.g., ``batch[i]``.
            grid: Sampling grid of image ``data`` oriented in world space.
                If ``None``, a default grid whose world space coordinate axes are aligned with the
                image axes, unit spacing, and origin at the image center is created on CPU.
            dtype: Data type of the image data. A copy of the data is only made when the desired
                ``dtype`` is not ``None`` and not the same as ``data.dtype``.
            device: Device on which to store image data. A copy of the data is only made when
                the data has to be copied to a different device.
            requires_grad: If autograd should record operations on the returned image tensor.
            pin_memory: If set, returned image tensor would be allocated in the pinned memory.
                Works only for CPU tensors.

        """
        # DataTensor.__new__() creates the tensor subclass given arguments:
        # data, dtype, device, requires_grad, pin_memory
        if self.ndim < 3:
            raise ValueError("Image tensor must have at least three dimensions (C, H, W)")
        self.grid_(grid)

    def _make_instance(
        self: TImage,
        data: Optional[Tensor] = None,
        grid: Optional[Grid] = None,
        **kwargs,
    ) -> TImage:
        r"""Create a new instance while preserving subclass meta-data."""
        kwargs["grid"] = self._grid if grid is None else grid
        return super()._make_instance(data, **kwargs)

    def __deepcopy__(self: TImage, memo) -> TImage:
        if id(self) in memo:
            return memo[id(self)]
        result = self._make_instance(
            self.data.clone(memory_format=torch.preserve_format),
            grid=self._grid.clone(),
            requires_grad=self.requires_grad,
            pin_memory=self.is_pinned(),
        )
        memo[id(self)] = result
        return result

    @staticmethod
    def _torch_function_grid(args) -> Optional[Grid]:
        r"""Get spatial sampling grid from args passed to __torch_function__."""
        if not args:
            return None
        if isinstance(args[0], (tuple, list)):
            args = args[0]
        grids: Sequence[Grid]
        grids = [g for g in (getattr(arg, "_grid", None) for arg in args) if g is not None]
        if not grids:
            return None
        return grids[0]

    @classmethod
    def _torch_function_result(cls, func, data, grid: Optional[Grid]) -> Any:
        if not isinstance(data, Tensor):
            return data
        if grid is not None and data.ndim == grid.ndim + 1 and data.shape[1:] == grid.shape:
            if func in (torch.clone, Tensor.clone):
                grid = grid.clone()
            if isinstance(data, cls):
                data._grid = grid
            else:
                data = cls(data, grid)
        elif type(data) != Tensor:
            data = data.as_subclass(Tensor)
        return data

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if func == F.grid_sample:
            raise ValueError("Argument of F.grid_sample() must be a batch, not a single image")
        data = Tensor.__torch_function__(func, (Tensor,), args, kwargs)
        grid = cls._torch_function_grid(args)
        if func in (
            torch.split,
            Tensor.split,
            torch.split_with_sizes,
            Tensor.split_with_sizes,
            torch.tensor_split,
            Tensor.tensor_split,
        ):
            return tuple(cls._torch_function_result(func, sub, grid) for sub in data)
        return cls._torch_function_result(func, data, grid)

    def batch(self: TImage) -> ImageBatch:
        r"""Image batch consisting of this image only.

        Because batched operations are generally more efficient, especially when executed on the GPU,
        most ``Image`` operations are implemented by ``ImageBatch``. The single-image batch instance
        property of this ``Image`` instance is used to execute single-image operations of ``self``.
        The ``ImageBatch`` uses a view on the tensor data of this ``Image``, as well as the ``Grid``
        object reference. No copies are made.

        """
        data = self.unsqueeze(0)
        grid = self._grid
        return ImageBatch(data, grid)

    def cube(self: TImage) -> Cube:
        r"""Get cube which defines the normalized coordinates space of the image with respect to the world."""
        return self._grid.cube()

    def domain(self: TImage) -> Domain:
        r"""Get oriented bounding box in world space which defines the domain within which the image is defined."""
        return self._grid.domain()

    def same_domain_as(self, other: Image) -> bool:
        """Check if this and another image have the same cube domain."""
        return self.same_domain_as(other.grid())

    @overload
    def grid(self: TImage) -> Grid:
        r"""Get sampling grid."""
        ...

    @overload
    def grid(self: TImage, grid: Grid) -> Image:
        r"""Get new image with given sampling grid."""
        ...

    def grid(self: TImage, grid: Optional[Grid] = None) -> Union[Grid, TImage]:
        r"""Get sampling grid or image with given grid, respectively."""
        if grid is None:
            return self._grid
        return self._make_instance(grid=grid)

    def grid_(self: TImage, grid: Optional[Grid]) -> TImage:
        r"""Change image sampling grid of this image."""
        if grid is None:
            grid = Grid(shape=self.shape[1:])
        elif grid.shape != self.shape[1:]:
            raise ValueError("Image grid size does not match spatial dimensions of image tensor")
        self._grid = grid
        return self

    def align_corners(self: TImage) -> bool:
        r"""Whether image resizing operations by default preserve corner points or grid extent."""
        return self._grid.align_corners()

    def center(self: TImage) -> Tensor:
        r"""Image center in world space coordinates as tensor of shape (D,)."""
        return self._grid.center()

    def origin(self: TImage) -> Tensor:
        r"""Image origin in world space coordinates as tensor of shape (D,)."""
        return self._grid.origin()

    def spacing(self: TImage) -> Tensor:
        r"""Image spacing in world units as tensor of shape (D,)."""
        return self._grid.spacing()

    def direction(self: TImage) -> Tensor:
        r"""Image direction cosines matrix as tensor of shape (D, D)."""
        return self._grid.direction()

    @property
    def sdim(self: TImage) -> int:
        r"""Number of spatial dimensions."""
        return self._grid.ndim

    @property
    def nchannels(self: TImage) -> int:
        r"""Number of image channels."""
        return self.shape[0]

    @classmethod
    def from_sitk(
        cls: Type[TImage],
        image: "sitk.Image",
        align_corners: bool = ALIGN_CORNERS,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ) -> TImage:
        r"""Create image from ``SimpleITK.Image`` instance."""
        try:
            from deepali.utils.simpleitk.torch import tensor_from_image
        except ImportError:
            raise RuntimeError(f"{cls.__name__}.from_sitk() requires SimpleITK")
        data = tensor_from_image(image, dtype=dtype, device=device)
        grid = Grid.from_sitk(image, align_corners=align_corners)
        return cls(data, grid)

    def sitk(self: TImage) -> "sitk.Image":
        r"""Create ``SimpleITK.Image`` from this image."""
        try:
            from deepali.utils.simpleitk.torch import image_from_tensor
        except ImportError:
            raise RuntimeError(f"{type(self).__name__}.sitk() requires SimpleITK")
        grid = self._grid
        origin = grid.origin().tolist()
        spacing = grid.spacing().tolist()
        direction = grid.direction().flatten().tolist()
        return image_from_tensor(self, origin=origin, spacing=spacing, direction=direction)

    @classmethod
    def from_uri(
        cls: Type[TImage],
        uri: str,
        align_corners: bool = ALIGN_CORNERS,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ) -> TImage:
        r"""Create image from data stored at a given URI."""
        return cls.read(uri, align_corners=align_corners, dtype=dtype, device=device)

    def to_uri(
        self: TImage,
        uri: str,
        compress: bool = True,
    ) -> TImage:
        r"""Save image data to file object at given URI."""
        self.write(uri, compress=compress)

    @classmethod
    def read(
        cls: Type[TImage],
        path: PathUri,
        align_corners: bool = ALIGN_CORNERS,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ) -> TImage:
        r"""Read image data from file."""
        data, grid = read_image(path)
        grid = grid.align_corners_(align_corners)
        return cls(data, grid, dtype=dtype, device=device)

    def write(self: TImage, path: PathUri, compress: bool = True) -> None:
        r"""Write image data to file."""
        write_image(self.tensor(), self.grid(), path, compress=compress)

    def normalize(
        self: TImage,
        mode: str = "unit",
        min: Optional[float] = None,
        max: Optional[float] = None,
    ) -> TImage:
        r"""Normalize image intensities in [min, max]."""
        batch = self.batch()
        batch = batch.normalize(mode=mode, min=min, max=max)
        return batch[0]

    def normalize_(
        self: TImage,
        mode: str = "unit",
        min: Optional[float] = None,
        max: Optional[float] = None,
    ) -> TImage:
        r"""Normalize image intensities in [min, max]."""
        batch = self.batch()
        batch = batch.normalize_(mode=mode, min=min, max=max)
        return batch[0]

    def rescale(
        self: TImage,
        min: Optional[Scalar] = None,
        max: Optional[Scalar] = None,
        data_min: Optional[Scalar] = None,
        data_max: Optional[Scalar] = None,
        dtype: Optional[DType] = None,
    ) -> TImage:
        r"""Clamp and linearly rescale image values."""
        batch = self.batch()
        batch = batch.rescale(min, max, data_min, data_max, dtype=dtype)
        return batch[0]

    def narrow(self: TImage, dim: int, start: int, length: int) -> TImage:
        r"""Narrow image along specified dimension."""
        batch = self.batch()
        batch = batch.narrow(dim + 1, start, length)
        return batch[0]

    def resize(
        self: TImage,
        size: Union[int, Array, Size],
        *args: int,
        mode: Union[Sampling, str] = Sampling.LINEAR,
        align_corners: Optional[bool] = None,
    ) -> TImage:
        r"""Interpolate image with specified spatial image grid size."""
        batch = self.batch()
        batch = batch.resize(size, *args, mode=mode, align_corners=align_corners)
        return batch[0]

    def resample(
        self: TImage,
        spacing: Union[float, Array, str],
        *args: float,
        mode: Union[Sampling, str] = Sampling.LINEAR,
    ) -> TImage:
        r"""Interpolate image with specified spacing."""
        batch = self.batch()
        batch = batch.resample(spacing, *args, mode=mode)
        return batch[0]

    def avg_pool(
        self: TImage,
        kernel_size: ScalarOrTuple[int],
        stride: Optional[ScalarOrTuple[int]] = None,
        padding: ScalarOrTuple[int] = 0,
        ceil_mode: bool = False,
        count_include_pad: bool = True,
    ) -> TImage:
        r"""Average pooling of image data."""
        batch = self.batch()
        batch = batch.avg_pool(
            kernel_size,
            stride=stride,
            padding=padding,
            ceil_mode=ceil_mode,
            count_include_pad=count_include_pad,
        )
        return batch[0]

    def downsample(
        self: TImage,
        levels: int = 1,
        dims: Optional[Sequence[SpatialDimArg]] = None,
        sigma: Optional[Union[Scalar, Array]] = None,
        mode: Optional[Union[Sampling, str]] = None,
        min_size: int = 0,
        align_corners: Optional[bool] = None,
    ) -> TImage:
        r"""Downsample image a given number of times."""
        batch = self.batch()
        batch = batch.downsample(
            levels,
            dims=dims,
            sigma=sigma,
            mode=mode,
            min_size=min_size,
            align_corners=align_corners,
        )
        return batch[0]

    def upsample(
        self: TImage,
        levels: int = 1,
        dims: Optional[Sequence[SpatialDimArg]] = None,
        sigma: Optional[Union[Scalar, Array]] = None,
        mode: Optional[Union[Sampling, str]] = None,
        align_corners: Optional[bool] = None,
    ) -> TImage:
        r"""Upsample image a given number of times."""
        batch = self.batch()
        batch = batch.upsample(
            levels, dims=dims, sigma=sigma, mode=mode, align_corners=align_corners
        )
        return batch[0]

    def pyramid(
        self: TImage,
        levels: int,
        start: int = 0,
        end: int = -1,
        dims: Optional[Sequence[SpatialDimArg]] = None,
        sigma: Optional[Union[Scalar, Array]] = None,
        mode: Optional[Union[Sampling, str]] = None,
        spacing: Optional[float] = None,
        min_size: int = 0,
        align_corners: Optional[bool] = None,
    ) -> Dict[int, TImage]:
        r"""Create Gaussian resolution pyramid."""
        batch = self.batch()
        batches = batch.pyramid(
            levels,
            start,
            end,
            dims=dims,
            sigma=sigma,
            mode=mode,
            spacing=spacing,
            min_size=min_size,
            align_corners=align_corners,
        )
        return {level: batch[0] for level, batch in batches.items()}

    def crop(
        self: TImage,
        margin: Optional[Union[int, Array]] = None,
        num: Optional[Union[int, Array]] = None,
        mode: Union[PaddingMode, str] = PaddingMode.CONSTANT,
        value: Scalar = 0,
    ) -> TImage:
        r"""Crop image at boundary."""
        batch = self.batch()
        batch = batch.crop(margin=margin, num=num, mode=mode, value=value)
        return batch[0]

    def pad(
        self: TImage,
        margin: Optional[Union[int, Array]] = None,
        num: Optional[Union[int, Array]] = None,
        mode: Union[PaddingMode, str] = PaddingMode.CONSTANT,
        value: Scalar = 0,
    ) -> TImage:
        r"""Pad image at boundary."""
        batch = self.batch()
        batch = batch.pad(margin=margin, num=num, mode=mode, value=value)
        return batch[0]

    def center_crop(self: TImage, size: Union[int, Array], *args: int) -> TImage:
        r"""Crop image to specified maximum size."""
        batch = self.batch()
        batch = batch.center_crop(size, *args)
        return batch[0]

    def center_pad(
        self: TImage,
        size: Union[int, Array],
        *args: int,
        mode: Union[PaddingMode, str] = PaddingMode.CONSTANT,
        value: Scalar = 0,
    ) -> TImage:
        r"""Pad image to specified minimum size."""
        batch = self.batch()
        batch = batch.center_pad(size, *args, mode=mode, value=value)
        return batch[0]

    def region_of_interest(self: TImage, *args, **kwargs) -> TImage:
        r"""Extract image region of interest."""
        batch = self.batch()
        batch = batch.region_of_interest(*args, **kwargs)
        return batch[0]

    def conv(
        self: TImage,
        kernel: Union[Tensor, Sequence[Optional[Tensor]]],
        padding: Union[PaddingMode, str, int] = None,
    ) -> TImage:
        r"""Filter image with a given (separable) kernel."""
        batch = self.batch()
        batch = batch.conv(kernel, padding=padding)
        return batch[0]

    @overload
    def sample(
        self: TImage,
        coords: Tensor,
        mode: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str, Scalar]] = None,
    ) -> Tensor:
        r"""Sample image at optionally deformed unit grid points.

        Note, to sample a set of 2D patches from a 3D volume, it may be beneficial to use ``ImageBatch.sample()``
        instead such that the output tensor shape is ``(N, C, Y, X)`` instead of ``(C, N, Y, X)`` given an
        input ``coords`` shape of ``(N, Y, X, D)``. For example, use ``image.batch().sample()``.

        Args:
            coords: Normalized coordinates of points at which to sample image as tensor of shape ``(..., D)``.
                Typical tensor shapes are: ``(Y, X, D)`` or ``(Z, Y, X, D)`` to sample an image at (deformed)
                2D or 3D grid points (cf. ``grid_sample()``), ``(N, D)`` to sample an image at a set of ``N``
                points, and ``(N, Y, X, D)`` to sample ``N`` 2D patches of size ``(X, Y)`` from a 3D volume.
            mode: Interpolation mode.
            padding: Extrapolation mode or scalar padding value.

        Returns:
            Tensor of sampled image values with shape ``(C, ...)``, where ``C`` is the number of channels
            of this image and ``...`` are the leading dimensions of ``coords`` (i.e., ``coords.shape[:-1]``).

        """
        ...

    @overload
    def sample(
        self: TImage,
        grid: Grid,
        mode: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str, Scalar]] = None,
    ) -> TImage:
        r"""Sample image at optionally deformed unit grid points.

        Args:
            grid: Sample this image at the points of the given sampling grid.
            mode: Interpolation mode.
            padding: Extrapolation mode or scalar padding value.

        Returns:
            Image sampled at grid points.

        """
        ...

    def sample(
        self: TImage,
        arg: Union[Grid, Tensor],
        mode: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str, Scalar]] = None,
    ) -> Union[Tensor, TImage]:
        r"""Sample image at points given as normalized coordinates.

        Note, to sample a set of 2D patches from a 3D volume, it may be beneficial to use ``ImageBatch.sample()``
        instead such that the output tensor shape is ``(N, C, Y, X)`` instead of ``(C, N, Y, X)`` given an
        input ``arg`` shape of ``(N, Y, X, D)``. For example, use ``image.batch().sample()``.

        Args:
            arg: Sampling grid defining points at which to sample image data, or normalized coordinates of
                points at which to sample image as tensor of shape ``(..., D)``. Typical tensor shapes are:
                ``(Y, X, D)`` or ``(Z, Y, X, D)`` to sample an image at (deformed) 2D or 3D grid points
                (cf. ``grid_sample()``), ``(N, D)`` to sample an image at a set of ``N`` points, and
                ``(N, Y, X, D)`` to sample ``N`` 2D patches of size ``(X, Y)`` from a 3D volume.
            mode: Interpolation mode.
            padding: Extrapolation mode or scalar padding value.

        Returns:
            If ``arg`` is of type ``Grid``, an ``Image`` with the sampled values and given sampling grid is returend.
            When ``arg == self.grid()``, a reference to ``self`` is returned. Otherwise, a ``Tensor`` of sampled image
            values with shape ``(C, ...)`` is returned, where ``C`` is the number of channels of this image and ``...``
            are the leading dimensions of ``grid`` (i.e., ``grid.shape[:-1]``).

        """
        batch = self.batch()
        if isinstance(arg, Grid):
            if arg == self.grid():
                return self
            batch = batch.sample(arg, mode=mode, padding=padding)
            assert isinstance(batch, ImageBatch)
            return batch[0]
        if isinstance(arg, Tensor):
            grid = arg.unsqueeze(0)
            data = batch.sample(grid, mode=mode, padding=padding)
            assert type(data) is Tensor
            assert data.shape[0] == 1
            data = data.squeeze(0)
            return data
        raise TypeError(f"{type(self).__name__}.sample() 'arg' must be Grid or Tensor")

    def __repr__(self) -> str:
        return type(self).__name__ + f"(data={self.tensor()!r}, grid={self.grid()!r})"

    def __str__(self) -> str:
        return type(self).__name__ + f"(data={self.tensor()!s}, grid={self.grid()!s})"



r"""Auxiliary types and functions to work with dataset partitions."""

from __future__ import annotations

from enum import Enum
from itertools import accumulate
from typing import List, Optional, Sequence, Tuple, Union

from torch import Generator, default_generator, randperm


__all__ = (
    "Partition",
    "dataset_split_lengths",
    "random_split_indices",
)


class Partition(Enum):
    r"""Enumeration of dataset partitions / splits."""

    NONE = "none"
    TEST = "test"  # use for testing a trained model
    TRAIN = "train"  # use for model training backward pass
    VALID = "valid"  # use for model training evaluation (validation)

    @classmethod
    def from_arg(cls, arg: Union[Partition, str, None]) -> Partition:
        r"""Create enumeration value from function argument."""
        if arg is None:
            return cls.NONE
        return cls(arg)


def dataset_split_lengths(
    total: int, ratios: Union[float, Sequence[float]]
) -> Tuple[int, int, int]:
    r"""Split dataset in training, validation, and test subset.

    The output ``lengths`` of this function can be passed to ``torch.utils.data.random_split`` to obtain
    the ``torch.utils.data.dataset.Subset`` for each split.

    See also:
        https://pytorch.org/docs/stable/data.html#torch.utils.data.random_split

    Args:
        total: Total number of samples in dataset.
        ratios: Fraction of samples in each split. When a float or 1-tuple is given,
            the specified fraction of samples is used for training and all remaining
            samples for validation during training. When a 2-tuple is given, the test
            set is assigned no samples. Otherwise, a 3-tuple consisting of ratios
            for training, validation, and test set, respectively should be given.
            The ratios must sum to one.

    Returns:
        lengths: Number of dataset samples in each subset.

    """
    if not isinstance(ratios, float) and len(ratios) == 1:
        ratios = ratios[0]
    if isinstance(ratios, float):
        ratios = (ratios, 1.0 - ratios)
    if len(ratios) == 2:
        ratios += (0.0,)
    elif len(ratios) != 3:
        raise ValueError(
            "dataset_split_lengths() 'ratios' must be float or tuple of length 1, 2, or 3"
        )
    if ratios[0] <= 0 or ratios[0] > 1:
        raise ValueError("dataset_split_lengths() training split ratio must be in (0, 1]")
    if any([ratio < 0 or ratio > 1 for ratio in ratios]):
        raise ValueError("dataset_split_lengths() ratios must be in [0, 1]")
    if sum(ratios) != 1:
        raise ValueError("dataset_split_lengths() 'ratios' must sum to one")
    lengths = [int(round(ratio * total)) for ratio in ratios]
    lengths[2] = max(0, lengths[2] + (total - sum(lengths)))
    lengths[1] = max(0, lengths[1] + (total - sum(lengths)))
    assert sum(lengths) == total
    return tuple(lengths)


def random_split_indices(
    lengths: Sequence[int],
    generator: Optional[Generator] = default_generator,
) -> List[List[int]]:
    r"""Randomly split dataset indices into non-overlapping sets of given lengths.

    Args:
        lengths: Lengths of splits to be produced.
        generator: Generator used for the random permutation.

    Returns:
        Lists of specified ``lengths`` with randomly selected indices in ``[0, sum(lengths))``.

    """
    subsets = []
    indices = randperm(sum(lengths), generator=generator).tolist()
    for offset, length in zip(accumulate(lengths), lengths):
        subsets.append(indices[offset - length : offset])
    return subsets
